;;
;@brief This file contains the basic math functions used by all the higher level functions.  
;
;  All functions in here only deal with the bare arrays nothing more.  The arrays are all structured with
;element 0 holding the used element count.  This is one of the resons why the sizeof operator may not give correct results (the other being if you use the internal memory management algorithm).
;  Almost all of the algs request a few extra bytes to hold the value, an extra for calculations that would create an overflow, plus the count element.  All of the memory requests go through the GetMem64
;alg, but depending on the value of the _usemalloc variable it will either use malloc or the internal memory management routines, and when it frees the memory, it allways calls Free64 and that routine
;will figure out whether to use the internal routines or free.  All of the algs get rid of leading zeros by adjusting the element count down till the element count is one (in which case the value of the array is zero).
;
;  There are special routines that are only meant to be used internally, the only functions meant to be used from outside of this file are the equality algs, the basic math algs, Mod, Pow, ModPow, ToString,
;the bitwise logical operators, the bitwise shifts, and the byte shifts.
;
;there is an explanation for each alg in the documentation.
;;

INCLUDE GlobalStuff64.inc
extrn malloc: PROC

	.data
		
	.code
	;;
	;@brief Checks for and eleminates leading zeros.
	;
	;This alg starts at the most significant byte of the array passed in and checks for zero, if it is zeros, it decrements the count by one, stores it in element 0, then continues the check.  This goes
	; on till either a non zero element is found or it reaches element 1.
	;@returns an array.
	;@param rcx - The array to check.
	;;
	ACheckForZeros PROC _val:PTR qword

		push	rdi
		mov		rdi, qword ptr[rcx]

CheckLoop:

		cmp		qword ptr[rcx+rdi*8], 0
		ja		CheckOut
		cmp		rdi, 1
		jna		CheckOut
		dec		rdi
		mov		qword ptr[rcx], rdi
		jmp		CheckLoop

CheckOut:

		pop		rdi
		mov		rax, rcx
		RET

	ACheckForZeros ENDP
	;;
	;@brief Counts how many digits an array would produce in a certain radix.
	;
	;This alg simply counts the digits  that would be created if the aray were converted to a string of a certain radix.
	;@returns a qword with the count.
	;@param rcx - The array to check.
	;;
	ACountDigits PROC _val:PTR qword, _tradix:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, qword ptr[rcx]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	malloc
		mov		_t2, rax
		mov		rdi, _t2
		mov		rsi, _v1
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		_var1, rax
		mov		qword ptr[rax], 1
		mov		rdx, _v2
		mov		qword ptr[rax+8], rdx
		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		_t3, rax
		mov		_t1, 0

MainLoop:

		mov		rcx, _t2
		cmp		qword ptr[rcx], 1
		ja		ItsGood
		cmp		qword ptr[rcx+8], 0
		je		GoOut

ItsGood:

		mov		rcx, _t2
		mov		rdx, _var1
		DivRem64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64
		mov		rcx, _t2
		call	Free64
		mov		rax, _t3
		mov		rax, qword ptr[rax]
		mov		_t2, rax
		inc		_t1
		jmp		MainLoop

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rcx, _t2
		call	Free64
		mov		rcx, _var1
		call	Free64
		mov		rcx, _t3
		call	Free64
		mov		rax, _t1
		cmp		rax, 0
		cmove	rax, _one

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	ACountDigits ENDP

	;This is the and function.  It does not reuire the 2 values to be the same length.  It takes the size of the shorter one, creates an array of that size
	;then moves that number of elements of _v1 over to it, lastly it ands that many elements of _v2 with the new array

	;;
	;@brief Does a bitwise and on 2 arrays.
	;
	;This alg simply and the elements of two arrays starting at the least significant byte till the smaller of the two arrays most significant byte is reached.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AAnd64 PROC _val1:PTR qword, _val2:PTR qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx					;move _val1 into the _v1 area
		mov		_v2, rdx					;move _val2 into the _v2 area
		mov		rax, qword ptr[rcx]			;move the element count into rax
		mov		rbx, 0						;clear rbx
		cmp		rax, qword ptr[rdx]			;compare element count in _v1 to the element count in _v2
		cmova	rax, qword ptr[rdx]			;if its above move the element count of _v2 into rax
		cmova	rbx, _one					;if its above move 1 into rbx 
		mov		_t1, rax					;move the smaller element count into _t1
		mov		_t2, rbx					;move the flag in rbx into _t2
		add		rax, 5						;make some head room in the new array
		mul		_eight						;multiply it by eight for qword
		mov		rcx, rax					;move the request amount into rcx
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0						;if rax is 0 then its an error
		je		MallocError					;jump to maloc error
		mov		_retval, rax				;otherwise move the pointer into _retval
		mov		r8, rax						;move the pointer into r8
		mov		rsi, _v1					;move _v1 into rsi
		mov		rdi, r8						;move tthe _retval into rdi
		mov		rcx, _t1					;move the smaller count into rcx
		mov		qword ptr[r8], rcx			;move the smaller count into r8
		add		rsi, 8						;adjust rsi past element 0
		add		rdi, 8						;adjust rdi past element 0
		pushf								;save the flags
		cld									;clear the direction flag
rep		movsq								;move the elements from _v1 to _retval
		popf								;restore the flags
		mov		rdi, 1						;point rdi to the first element
		mov		rcx, _v2					;move _v2 into rcx

AndLoop:

		mov		rax, qword ptr[rcx+rdi*8]	;move the element from _v2 into rax
		and		qword ptr[r8+rdi*8], rax	;and rax to the element in r8
		inc		rdi							;inc rdi
		cmp		rdi, _t1					;compare it to the smaller count
		jna		AndLoop						;if its not above jump to AndLoop

MallocError:

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		
		mov		rax, r8						;move _retval into rax
		RET

	AAnd64 ENDP

	;this is the or function, it finds the larger of the two arrays and creates an array with that many elements in it.  It then copys the _v1 val over to the new array
	;if _v1 was the smaller array then it zeros out all the elements over what _v1 has.  Lastly it ors each element with _v2, if _v2 was the smaller it will continue
	;oring the answer array with 0

	;;
	;@brief Does a bitwise or on 2 arrays.
	;
	;This alg ors the elements of the 2 arrays starting at the least significant byte till the smaller of the two arrays is reached, then the larger is ord with 0 till its most significant
	;byte is reached.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AOr64 PROC _val1:PTR qword, _val2:PTR qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, qword ptr[rcx]
		mov		rbx, 0
		cmp		rax, qword ptr[rdx]
		cmovb	rax, qword ptr[rdx]
		cmovb	rbx, _one
		cmovb	rcx, qword ptr[rcx]
		cmovnb	rcx, qword ptr[rdx]
		mov		_t1, rax
		mov		_t2, rbx
		mov		_t3, rcx
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		_retval, rax
		mov		r8, rax
		mov		rsi, _v1
		mov		rdi, r8
		mov		rcx, _t1
		mov		qword ptr[r8], rcx
		add		rsi, 8
		add		rdi, 8
		pushf
		cld
rep		movsq
		popf
		cmp		_t2, 0
		je		PrepOrLoop
		mov		rdx, _v1
		mov		rcx, _v2
		mov		rcx, qword ptr[rcx]
		sub		rcx, qword ptr[rdx]
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf

PrepOrLoop:

		mov		rdi, 1
		mov		rcx, _v2

OrLoop:

		mov		rax, qword ptr[rcx+rdi*8]
		xor		qword ptr[r8+rdi*8], rax
		inc		rdi
		cmp		rdi, qword ptr[rcx]
		jna		OrLoop	
		
MallocError:

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		
		mov		rax, r8
		RET

	AOr64 ENDP

	;this is the exclusive or function, it finds the larger of the two arrays, makes an array of that size, then copies _v1 to it, if _v1 was the smaller array, it zeros out the rest of the elements.
	;then it xors _v2 with it, if _v2 was the smaller array it xors the rest with zero.

	;;
	;@brief Does a bitwise exclusive or on 2 arrays.
	;
	;This alg exclusive ors each element of 2 arrays starting at the least significant byte till the most significant byte of the smaller array is reached, then it continues by exclusive oring the
	;lasrger array with 0 till its most significant byte is reached.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AXOr64 PROC _val1:PTR qword, _val2:PTR qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, qword ptr[rcx]
		mov		rbx, 0
		cmp		rax, qword ptr[rdx]
		cmovb	rax, qword ptr[rdx]
		cmovb	rbx, _one
		mov		_t1, rax
		mov		_t2, rbx
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		_retval, rax
		mov		r8, rax
		mov		rsi, _v1
		mov		rdi, r8
		mov		rcx, _t1
		mov		qword ptr[r8], rcx
		add		rsi, 8
		add		rdi, 8
		pushf
		cld
rep		movsq
		popf
		cmp		_t2, 0
		je		PrepOrLoop
		mov		rdx, _v1
		mov		rcx, _v2
		mov		rcx, qword ptr[rcx]
		sub		rcx, qword ptr[rdx]
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf

PrepOrLoop:

		mov		rdi, 1
		mov		rcx, _v2

XOrLoop:

		mov		rax, qword ptr[rcx+rdi*8]
		xor		qword ptr[r8+rdi*8], rax
		inc		rdi
		cmp		rdi, qword ptr[rcx]
		jna		XOrLoop	
		
XorLoop2:	

		cmp		rdi, _t3
		ja		GoOut
		xor		qword ptr[r8+rdi*8], 0
		inc		rdi
		jmp		XOrLoop2

GoOut:

MallocError:

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		
		mov		rax, r8
		RET

	AXOr64 ENDP

		;the equality algorithms work under the assumption that all may be worked from implementing just 2 of them to make it easier to maintain
		;so we write out algorithms for the Equals and the LessThan all other algs are worked through these
		;AE uses the Equals, ANE uses equals then nots the result, LT uses LessThan, GT uses LessThan with the inputs reversed, 
		;LTOE uses the LessThan with the inputs reversed and not the result and GTOE uses the LessThan and not the result

	;;
	;@brief Checks for equality between 2 arrays.
	;
	;This alg simply compares each element of 2 arrays for equality.
	;@returns a bool.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AE64 PROC _val1:PTR qword, _val2:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rdi, rcx
		mov		rsi, rdx
		mov		rcx, qword ptr[rdi]
		inc		rcx
		pushf
		std
repe	cmpsq
		jne		eNotEqual
		popf
		mov		rax, 1
		jmp		eGoOut
		
eNotEqual:

		popf
		mov		rax, 0

eGoOut:

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	AE64 ENDP

		;not equal algorithm uses the equals algorithm then nots it

	;;
	;@brief Checks for inequality between 2 arrays.
	;
	;This alg sends the 2 arrays to the AE alg then nots the result.
	;@returns a bool.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	ANE64 PROC neval1offset:PTR qword, neval2offset:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		E64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		
neAllreadySet:	

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	ANE64 ENDP	

		;the less than algorithm
	
	;;
	;@brief Checks if the first array is less than the second array.
	;
	;This alg compares each element, starting at the most significant byte till it finds 2 elements that do not equal each other or it reaches the end of the array.
	;@returns a bool.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	ALT64 PROC ltval1offset:PTR qword, ltval2offset:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rsi, rcx
		mov		rdi, rdx
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		std
repe	cmpsq
		jl		ltLessThan
		popf
		mov		rax, 0
		jmp		ltGoOut

ltNotLessThan:

		mov		rax, 0
		jmp		ltGoOut

ltLessThan:

		popf
		mov		rax, 1

ltGoOut:

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
			
	ALT64 ENDP	

		;less than or equals reverses the values calls the less than algorithm then nots the answer

	;;
	;@brief Checks if the first array is less than or equal to the second array.
	;
	;This alg reverses the arrays, sends them to the less than alg, and nots the result.
	;@returns a bool.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	ALTOE64 PROC ltoeval1offset:PTR qword, ltoeval2offset:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rax, rcx
		mov		rcx, rdx
		mov		rdx, rax
		LT64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		
ltoeAllreadySet:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	ALTOE64 ENDP

		;greater than switches the values then calls the less than algorithm

	;;
	;@brief Checks if the first array is greater than the second array.
	;
	;This alg reverses the arrays then sends them to the less than alg.
	;@returns a bool.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AGT64 PROC gtval1offset:PTR qword, gtval2offset:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rax, rcx
		mov		rcx, rdx
		mov		rdx, rax
		LT64		

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	AGT64 ENDP

		;greater than or equaal calls the less than algorithm then nots the answer

	;;
	;@brief Checks if the first array is greater than or equal to the second array.
	;
	;This alg sends the arrays to the less than alg then nots the result.
	;@returns a bool.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AGTOE64 PROC gtoeval1offset:PTR qword, gtoeval2offset:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		LT64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		
gtoeAllreadySet:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	AGTOE64 ENDP
	;;
	;@brief Adds 2 arrays.
	;
	;This alg adds each element of the arrays starting at the least significant bytes, till the most significant byte of the smaller array is reached, then it adds each element of the larger array to 0
	;plus the carry till its most significant byte is reached.  Then it sends the result to the CheckForZeros alg.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AAdd642 PROC aval1:PTR qword, aval2:PTR qword
	
		;The add alg is pretty standard, find the size of the larger array and create and copy the larger array _one larger for the result
		;then add the smaller val to it cell by cell till you reach the final cell in the smaller array
		;Then the alg will adc the result array until there carry bit clears
		;The final part checks to make sure n_one of the high order bytes are _zero, if they are it dec the array size till it hit a non-_zero cell

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rdi, qword ptr[r12]
		mov		rax, qword ptr[r11]
		cmp		rax, qword ptr[r12]
		cmovb	rax, qword ptr[r12]
		cmovb	rdi, qword ptr[r11]
		cmovb	r11, rdx
		cmovb	r12, rcx
		cmove	rdx, _two
		cmova	rdx, _zero
		mov		_t5, rdx
		mov		_t1, rdi
		mov		_jfunc, rax
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		mov		rax, _jfunc
		cmp		r8, 0
		je		MallocError
		mov		rdi, r8
		mov		rsi, r11
		mov		rcx, qword ptr[r11]
		mov		qword ptr[r8+rcx*8+8], 0
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		inc		qword ptr[r8]
		mov		rdi, 1
		xor		rdx, rdx

AddLoop:

		mov		rax, qword ptr[r12+rdi*8]
		add		qword ptr[r8+rdi*8], rax
		cmovc	rdx, _one
		inc		rdi
		add		qword ptr[r8+rdi*8], rdx
		cmovnc	rdx, _zero
		cmp		rdi, _t1
		jna		AddLoop
		cmp		_t5, 1
		ja		CheckForZs

AddLoop2:

		cmp		rdx, 0
		je		CheckForZs
		inc		qword ptr[r8+rdi*8]
		cmovnc	rdx, _zero
		inc		rdi
		cmp		rdi, qword ptr[r8]
		jb		AddLoop2

CheckForZs:

		mov		rcx, r8
		CHECKFORZEROS
		mov		r8, rax
		jmp		GoOut

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		ErrorOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

ErrorOut:

		mov		rax, rax

GoOut:

		mov		rax, r8

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	AAdd642 ENDP
	;;
	;@brief Adds 2 arrays.
	;
	;This alg adds each element of the arrays starting at the least significant bytes, till the most significant byte of the smaller array is reached, then it adds each element of the larger array to 0
	;plus the carry till its most significant byte is reached.  Then it sends the result to the CheckForZeros alg.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AAdd64 PROC aval1:PTR qword, aval2:PTR qword
	
		;The add alg is pretty standard, find the size of the larger array and create and copy the larger array _one larger for the result
		;then add the smaller val to it cell by cell till you reach the final cell in the smaller array
		;Then the alg will adc the result array until there carry bit clears
		;The final part checks to make sure n_one of the high order bytes are _zero, if they are it dec the array size till it hit a non-_zero cell

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, qword ptr[r11]
		cmp		rax, qword ptr[r12]
		ja		V1Big
		jb		V2Big
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		rdi, r8
		mov		rsi, _v1
		mov		rcx, qword ptr[rsi]
		inc		rcx
		mov		qword ptr[r8+rcx*8], 0
		pushf
		cld
rep		movsq
		popf
		mov		r11, _v2
		mov		rdi, 1
		mov		rcx, 0
		
AddLoopE:

		add		qword ptr[r8+rdi*8], rcx
		cmovc	rcx, _one
		cmovnc	rcx, _zero
		mov		rax, qword ptr[r11+rdi*8]
		add		qword ptr[r8+rdi*8], rax
		cmovc	rcx, _one
		cmovnc	rcx, _zero
		inc		rdi
		cmp		rdi, qword ptr[r11]
		jna		AddLoopE
		cmp		rcx, 0
		je		GoOut
		inc		qword ptr[r8]
		mov		qword ptr[r8+rdi*8], 1
		jmp		GoOut
		
V1Big:

		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		rdi, rax
		mov		rsi, _v1
		mov		rcx, qword ptr[rsi]
		inc		rcx
		mov		qword ptr[r8+rcx*8], 0
		pushf
		cld
rep		movsq
		popf
		mov		r11, _v2
		mov		rdi, 1
		mov		rcx, 0
		
AddLoopV1:

		add		qword ptr[r8+rdi*8], rcx
		cmovc	rcx, _one
		cmovnc	rcx, _zero
		mov		rax, qword ptr[r11+rdi*8]
		add		qword ptr[r8+rdi*8], rax
		cmovc	rcx, _one
		cmovnc	rcx, _zero
		inc		rdi
		cmp		rdi, qword ptr[r11]		
		jna		AddLoopV1

AddLoopV12:

		cmp		rcx, 0
		je		GoOut
		add		qword ptr[r8+rdi*8], 1
		cmovnc	rcx, _zero
		inc		rdi
		cmp		rdi, qword ptr[r8]
		jna		AddLoopV12
		cmp		rcx, 0
		je		GoOut
		inc		qword ptr[r8]
		mov		qword ptr[r8+rdi*8], 1
		jmp		GoOut
		
V2Big:

		mov		rax, qword ptr[r12]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		rdi, rax
		mov		rsi, _v2
		mov		rcx, qword ptr[rsi]
		inc		rcx
		mov		qword ptr[r8+rcx*8], 0
		pushf
		cld
rep		movsq
		popf
		mov		r11, _v1
		mov		rdi, 1
		mov		rcx, 0
		
AddLoopV2:

		add		qword ptr[r8+rdi*8], rcx
		cmovc	rcx, _one
		cmovnc	rcx, _zero
		mov		rax, qword ptr[r11+rdi*8]
		add		qword ptr[r8+rdi*8], rax
		cmovc	rcx, _one
		cmovnc	rcx, _zero
		inc		rdi
		cmp		rdi, qword ptr[r11]		
		jna		AddLoopV2

AddLoopV22:

		cmp		rcx, 0
		je		GoOut
		add		qword ptr[r8+rdi*8], 1
		cmovnc	rcx, _zero
		inc		rdi
		cmp		rdi, qword ptr[r8]
		jna		AddLoopV22
		cmp		rcx, 0
		je		GoOut
		inc		qword ptr[r8]
		mov		qword ptr[r8+rdi*8], 1
		jmp		GoOut

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		ErrorOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

ErrorOut:

		mov		rax, rax

GoOut:

		mov		rcx, r8
		CHECKFORZEROS
		mov		r8, rax
		mov		rax, r8

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	AAdd64 ENDP
	;;
	;@brief Subtracts 2 arrays.
	;
	;This is the only alg that has any requirements as to the order of the arrays.  In this alg the first array has to be larger then the second array.  It makes a copy of _val1, then starting at the
	;least significant byte, subtracts the elements till the end of the second array is reached, then it subtracts the carry from each element till the end of the larger array is reached.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	ASub64 PROC sval1:PTR qword, sval2:PTR qword
	
		;This alg is a standard subtraction alg
		;it creates an answer array the size of _val1
		;then subtracts the arrays cell by cell till it reaches the final cell of the smaller array
		;at which time it continues doing subtractions with carry on the larger till all of the bytes are transfered
		;the final part does the _zero check

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		rcx, rdx
		mov		rdx, r11

SwitchEm:

		mov		r11, rcx
		mov		rcx, rdx
		mov		rdx, r11
		mov		r11, rcx
		mov		r12, rdx
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r12]
		ja		ItsAllGood
		jb		SwitchEm
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rdi, r12
		mov		rsi, r11
		add		rdi, rax
		add		rsi, rax
		pushf
		std
repe	cmpsq
		ja		PrepItsAllGood
		jb		PrepSwitchEm
		popf
		jmp		ReturnZero

PrepItsAllGood:

		popf
		ja		ItsAllGood

PrepSwitchEm:

		popf
		jb		SwitchEm

ItsAllGood:

		mov		rax, qword ptr[r11]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		cmp		r8, 0
		je		MallocError
		mov		rdi, 1
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		inc		 rcx
		mov		qword ptr[r8+rcx*8], 0
		pushf
		cld
rep		movsq
		popf
		inc		qword ptr[r8]
		mov		rdi, 1
		xor		rdx, rdx

SubLoop:

		mov		rax, qword ptr[r12+rdi*8]
		sub		qword ptr[r8+rdi*8], rax
		cmovc	rdx, _one
		inc		rdi
		sub		qword ptr[r8+rdi*8], rdx
		cmovnc	rdx, _zero
		cmp		rdi, qword ptr[r12]
		jna		SubLoop

CarryLoop:

		inc		rdi
		cmp		rdi, qword ptr[r8]
		jnb		CheckForZs
		cmp		rdx, 0
		je		CheckForZs
		sub		qword ptr[r8+rdi*8], rdx
		cmovnc	rdx, _zero
		jmp		CarryLoop

SubLoop2:

		mov		rax, qword ptr[r11+rdi*8]
		sub		rax, rdx
		cmovc	rdx, _one
		cmovnc	rdx, _zero
		mov		qword ptr[r8+rdi*8], rax
		inc		rdi
		cmp		rdi, qword ptr[r11]
		jna		SubLoop2

CheckForZs:

		mov		rcx, r8
		CHECKFORZEROS
		mov		r8, rax
		jmp		GoOut

ReturnZero:

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		r8, rax
		jmp		GoOut

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		ErrorOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

ErrorOut:

		mov		rax, rax

GoOut:

		mov		rax, r8

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET
	
	ASub64 ENDP
	;;
	;@brief Multiplies 2 arrays.
	;This is a trial alg to see if there is an improvement to the multiply alg by using shifts and adds.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AMultiply642 PROC mval1:PTR qword, mval2:PTR qword

		;this version of the multiply algorithm uses the shift and add method.  We start by creating an array large enough to hold the result
		;and initialize it to zero.  Then we will shift through r12 starting at the most significant byte one bit at a time counting zeros
		;when we hit a set bit we jump to the shifting section to shift the result (the entire array being used) by however many bits we just calculated.
		;after the shift we proceed to add r11 into the result, then repeat till the end of r12 is reached.  After the last shift there may be no need for a final add
		;this depends on whether bit 0 of the least significant byte of r12 is set or not.

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _one, _zero, _one, _zero

		;we enter the alg, find the larger of the two values and make sure it is in r11, the smaller goes in r12.  We will use r11 for the additions, and r12
		;will be what we shift to find the adds for the result.

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_t2, 0
		mov		r11, rcx
		mov		r12, rdx
		mov		rcx, qword ptr[r11]
		cmp		rcx, qword ptr[r12]
		cmova	r12, r11
		cmova	r11, rdx
		jne		MoveOn
		mov		rax, rcx
		mul		_eight
		mov		rdi, r11
		mov		rsi, r12
		add		rdi, rax
		add		rsi, rax
		pushf
		std
repe	cmpsq
		cmova	rax, r12
		cmova	r12, r11
		cmova	r11, rax
		popf

MoveOn:

		;here we create the array to hold the result, initialize it to zero, and set the first element to one

		mov		rax, qword ptr[r11]
		add		rax, qword ptr[r12]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		rdi, rax
		add		rdi, 8
		mov		r8, rax
		mov		rcx, qword ptr[r11]
		add		rcx, qword ptr[r12]
		inc		rcx
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf
		mov		qword ptr[r8], 1
		mov		rdi, qword ptr[r12]
		inc		rdi

FindByteLoop:

		;this section simply checks to make sure the most significant byte does not contain zero

		cmp		rdi, 1
		jna		GoOut
		dec		rdi
		cmp		qword ptr[r12+rdi*8], 0
		je		FindByteLoop
		mov		_t3, rdi
		mov		r15, qword ptr[r12+rdi*8]
		mov		r13, 64

MainLoop:

		; this is the main loop for the alg.  Here we shift r15 one bit at a time till we reach the end of the bytes in r12 or find a set bit.
		;then we jump to check if a shift needs to be done, and if an addition needs to be done.

		mov		rcx, 1							;this is for the bit shift of the byte in r15
		xor		rax, rax						;this is to keep track of how many shifts to do this set
		xor		r9, r9							;this will signal that a shift needs to be done
		xor		r10, r10						;this will signal an add needs to be done

FindBitLoop:

		;this section shifts r15 to the left one bit at a time till a set bit is found (the carry flag will be set after the shift).  If r13 reaches zero, then we have cycled through all the bits
		;and we grab another byte from r12 if there is one, if not it will jump to the check to see if a shift needs to be done.  The check is done through r9
		;which is set when a shift of r15 happens, (if no shift of r15, then no shift of our result).  If a set bit is found, then r10 is set to one
		;to signal that we need to do an add.

		cmp		r13, 0
		ja		StillBits
		mov		rdi, _t3
		cmp		rdi, 1
		jna		CheckShift
		dec		rdi
		mov		_t3, rdi
		mov		r15, qword ptr[r12+rdi*8]
		mov		r13, 64

StillBits:

		mov		r9, 1
		dec		r13
		inc		rax
		sal		r15, cl
		cmovc	r10, _one
		jnc		FindBitLoop

CheckShift:

		;start by checking r9.  The last shift required may allready have executed, if it has then exit.
		;in this block we will prep for the shift by first dividing the number of shifts calcd by 64, if rax > 0 then we can shift entire bytes to the left
		;to speed it up some.  if rax==0 then jump to the bit shifts

		cmp		r9, 0
		je		CheckAdd
		xor		rdx, rdx
		div		_sixtyfour
		mov		_t1, rax						;this holds the number of bytes to shift to the left
		mov		_t2, rdx						;this holds the number of bits to shift to the left
		cmp		rax, 0
		je		NoBytes
		mov		rax, qword ptr[r8]
		mov		rcx, rax
		mul		_eight
		mov		rsi, r8
		add		rsi, rax
		mov		rax, _t1
		mul		_eight
		mov		rdi, rsi
		add		rdi, rax
		pushf
		std
rep		movsq
		mov		rcx, _t1
		mov		rdi, r8
		add		rdi, 8
		mov		rax, 0
rep		stosq
		popf
		mov		rax, qword ptr[r8]
		add		rax, _t1
		mov		qword ptr[r8], rax

NoBytes:

		;this section is where we start prepping for any bit shifts, we start by checking if _t2 is zero, if it is there are no bit shifts

		cmp		_t2, 0
		je		CheckAdd
		mov		rdi, qword ptr[r8]
		inc		rdi

CheckThisByte:

		;now we find the most significant bit in the most significant byte, if that plus the number of shifts > 64 then we need to inc the number of elements in the array

		cmp		rdi, 1
		jna		CheckAdd
		dec		rdi
		cmp		qword ptr[r8+rdi*8], 0
		je		CheckThisByte
		bsr		rcx, qword ptr[r8+rdi*8]
		inc		rcx
		add		rcx, _t2
		mov		rsi, rdi
		cmp		rcx, 64
		jna		DontInc
		inc		rdi
		mov		qword ptr[r8], rdi
		dec		rdi
		
DontInc:

		;here we prepare the number of right shifts to left shifts.  

		mov		rbx, _t2
		mov		rcx, 64
		sub		rcx, rbx
		mov		rsi, rdi
		inc		rdi

ShiftLoop:

		;The way this part of the algorithm works is that it will move the byte into rax
		;where it will get shifted right 64 - number of bits left we are doing.  This will be the bits getting added to the byte above it (the bits getting shifted out of this byte into the next)
		;then we shift the memory location holding the element to the left (leaving the lsb bits zerod)

		mov		rax, qword ptr[r8+rsi*8]
		shr		rax, cl
		add		qword ptr[r8+rdi*8], rax
		xchg	rbx, rcx
		sal		qword ptr[r8+rsi*8], cl
		xchg	rbx, rcx
		dec		rdi
		dec		rsi
		cmp		rsi, 0
		jne		ShiftLoop
		
CheckAdd:

		;here we are checking if an add needs to be done (did we hit a set bit or are we here because we reached the end of r12 and bit 0 == 0

		cmp		r10, 0
		je		NoAdd
		mov		rdi, 1
		mov		rbx, 0

AddLoop1:

		;standard add loop, mov the element in the r11 array into rax, add this to the element in the result array, if there is a carry then add
		;it to the next higher bit in the result array.  Continue till we reach the end of the r11 array

		mov		rax, qword ptr[r11+rdi*8]
		add		qword ptr[r8+rdi*8], rax
		cmovc	rbx, _one
		inc		rdi
		cmp		rbx, 0
		je		NoCarryHere
		add		qword ptr[r8+rdi*8], rbx
		cmovnc	rbx, _zero
		cmp		rdi, qword ptr[r8]
		jna		NoCarryHere
		mov		qword ptr[r8], rdi

NoCarryHere:

		cmp		rdi, qword ptr[r11]
		jna		AddLoop1

AddLoop2:

		;this part of the loop is just taking care of the carry from the previous addition, we continue this till there is no carry

		cmp		rbx, 0
		je		AddOut
		cmp		rdi, qword ptr[r8]
		ja		AddOut
		add		qword ptr[r8+rdi*8], rbx
		cmovnc	rbx, _zero
		inc		rdi
		jmp		AddLoop2

AddOut:

		cmp		rbx, 0
		je		NoAdd
		inc		qword ptr[r8]
		mov		qword ptr[r8+rdi*8], 1

NoAdd:

		;prep to check the most significant byte is zero

		mov		rcx, r8
		CHECKFORZEROS
		mov		r8, rax
		cmp		r9, 0
		ja		MainLoop
		cmp		r10, 0
		ja		MainLoop

GoOut:

		;mov the rsult into rax, reset the stack pointer, and pop our saved values then return

		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _one, _zero, _one, _zero

		RET

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		ErrorOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

ErrorOut:

		jmp		GoOut
				
ReturnZero:

		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0

		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _one, _zero, _one, _zero

		RET
	
	AMultiply642 ENDP
	;;
	;@brief Multiplies 2 arrays.
	;
	;This is the standard multiply alg.  It creates an array and zeros out all the elements, then it multiplies each element of each array and adds it in the appropriate place in the result array.
	;  Then it sends the result to the checkforzeros routine.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AMultiply64 PROC mval1:PTR qword, mval2:PTR qword

		;This is a standard multiply alg, it creates an answer array the size of _v1size+_v2size+1 and clears all the cells
		;starting at cell 1 of _one of (stationary) the arrays the alg will multiply that cell by each cell in the other array and adding it to the (_currentcellv1-1)+(_currentcellv2-1)+1 in the answer array
		;at the end of a cycle it takes any carry from the last multiply add round and stores it in the next cell of the answer array
		;the last thing that is done is to check the high order bytes for _zero

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, qword ptr[r11]		
		add		rax, qword ptr[r12]				
		inc		rax	
		mov		_jfunc, rax
		inc		rax						
		mul		_eight					
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax	
		mov		rax, 	_jfunc
		cmp		r8, 0
		je		MallocError					
		mov		qword ptr[r8], rax
		mov		rcx, rax
		mov		rdi, r8
		add		rdi, 8
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf
		mov		rdi, 1
		mov		rsi, 1
		mov		rcx, 1
		mov		rdx, 0

MultLoop:

		mov		rbx, rdx
		mov		rax, qword ptr[r11+rdi*8]
		mul		qword ptr[r12+rsi*8]
		add		qword ptr[r8+rcx*8], rax
		adc		rdx, 0
		add		qword ptr[r8+rcx*8], rbx
		adc		rdx, 0
		inc		rcx
		inc		rdi
		cmp		rdi, qword ptr[r11]
		jna		MultLoop
		mov		qword ptr[r8+rcx*8], rdx
		mov		rdx, 0
		mov		rdi, 1
		inc		rsi
		mov		rcx, rsi
		cmp		rsi, qword ptr[r12]
		jna		MultLoop
		jmp		GoOut

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		ErrorOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

ErrorOut:

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		rax, 0
		RET

GoOut:

		mov		rcx, r8
		CHECKFORZEROS

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	AMultiply64 ENDP
	;;
	;@brief Squares an array.
	;
	;This is basically the same alg as the multiply routine except there is only one input.
	;@returns an array.
	;@param rcx - The array to square.
	;;
	ASquare64 PROC val1:PTR qword

		;thiis is a modified version of the multiply alg

		ENTERFUNC _zero, _one, _zero, _one, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		rax, qword ptr[r11]
		add		rax, rax
		inc		rax
		mov		_jfunc, rax
		add		rax, 7
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		mov		rax, _jfunc
		cmp		r8, 0
		je		MallocError
		dec		rax
		mov		qword ptr[r8], rax
		mov		rdi, qword ptr[r11]
		cmp		rdi, 1
		ja		Cont
		mov		rax, qword ptr[r11+8]
		cmp		rax, 0
		je		ReturnZero
		cmp		rax, 1
		je		ReturnOne
		mul		rax
		mov		qword ptr[r8+8], rax
		mov		qword ptr[r8], 1
		cmp		rdx, 0
		je		GoOut
		mov		qword ptr[r8+16], rdx
		mov		qword ptr[r8], 2
		jmp		GoOut

Cont:

		mov		rcx, qword ptr[r8]
		mov		rdi, r8
		add		rdi, 8
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf
		mov		rdi, 1
		mov		rsi, 1
		mov		rcx, 1
		mov		rdx, 0

MultLoop:

		mov		rbx, rdx
		mov		rax, qword ptr[r11+rdi*8]
		mul		qword ptr[r11+rsi*8]
		add		rax, rbx
		adc		rdx, 0
		add		qword ptr[r8+rcx*8], rax
		adc		rdx, 0
		inc		rdi
		inc		rcx
		cmp		rdi, qword ptr[r11]
		jna		MultLoop
		mov		qword ptr[r8+rcx*8], rdx
		mov		rdx, 0
		mov		rdi, 1
		inc		rsi
		mov		rcx, rsi
		cmp		rsi, qword ptr[r11]
		jna		MultLoop
		jmp		GoOut

ReturnZero:

		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		jmp		OutOfHere

ReturnOne:

		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 1
		jmp		OutOfHere

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rcx, r8
		CHECKFORZEROS
		mov		r8, rax

OutOfHere:

		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _one, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	ASquare64 ENDP
	;;
	;@This is a special version of the multiply routine used in the divide and mod routines.
	;
	;This is basically the same alg as the multiply routine except instead of creating a return array it uses the temp array from the divide/mod algs
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	ADivMult64 PROC l1:PTR qword, _val2:qword

		;this is basically the same as the multiply alg except val2 is a single qword value mainly for the divide alg

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rdx
		mov		rdi, 1
		mov		rdi, 1
		mov		rdx, 0
		mov		rax, qword ptr[r12]
		mov		qword ptr[r9], rax

MainLoop:

		mov		rbx, rdx
		mov		rax, qword ptr[r12+rdi*8]
		mul		r11
		add		rax, rbx
		adc		rdx, 0
		mov		qword ptr[r9+rdi*8], rax
		inc		rdi
		cmp		rdi, qword ptr[r12]
		jna		MainLoop
		cmp		rdx, 0
		je		GoOut
		mov		qword ptr[r9+rdi*8], rdx
		mov		qword ptr[r9], rdi
		jmp		GoOut

GoOut:

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	ADivMult64 ENDP
	;;
	;@This is a special version of the subtract routine used in the divide and mod routines.
	;
	;This is basically the same alg as the subtract routine except instead of creating a return array it uses the temp array from the divide/mod algs
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	ADivSub64 PROC sval1:PTR qword, sval2:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx

		;the managed code is supposed to have made sure that _val1>=_val2 so we know the return value will be no bigger than _val1

		mov		rdi, 1
		mov		rax, qword ptr[r11]
		xor		rdx, rdx

SubLoop:

		;here is the main sub loop storing any borrows from the previous loop in rdx

		mov		rcx, rdx
		mov		rax, qword ptr[r9+rdi*8]
		sub		qword ptr[r11+rdi*8], rax
		cmovc	rdx, _one
		cmovnc	rdx, _zero
		sub		qword ptr[r11+rdi*8], rcx
		cmovc	rdx, _one
		inc		rdi
		cmp		rdi, qword ptr[r9]
		jna		SubLoop

SubLoop2:

		;if the length of _val1 is greater than the length of _val2 then the alg continues here basically copying the elements in _val1 with respect to borrows

		cmp		rdi, qword ptr[r11]
		ja		CheckForZs
		sub		qword ptr[r11+rdi*8], rdx
		cmovc	rdx, _one
		cmovnc	rdx, _zero
		inc		rdi
		jmp		SubLoop2

CheckForZs:

		mov		rcx, r11
		CHECKFORZEROS
		mov		r11, rax
		jmp		GoOut

GoOut:

		mov		rax, r11

		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
	
	ADivSub64 ENDP
	;;
	;@This is a special version of the BitShiftLeft routine used in the divide and mod routines.
	;
	;This is basically the same alg as the BitShiftLeft routine except instead of creating a return array it uses the temp array from the divide/mod algs
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	ADivBitShiftLeft64 PROC val:PTR qword, count:qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, r12
		xor		rdx, rdx
		div		_sixtyfour
		mov		_t1, rax
		cmp		rdx, 0
		ja		DoBits
		mov		rcx, r11
		mov		rdx, rax
		mov		rax, r8
		call	ADivByteShiftLeft64

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

DoBits:

		mov		_t2, rdx
		mov		rax, qword ptr[r11]
		add		rax, _t1
		inc		rax
		mov		qword ptr[r9], rax
		mov		rcx, rax
		mov		rdi, r9
		add		rdi, 8
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf
		mov		rdi, 1
		mov		rsi, 1
		add		rsi, _t1
		mov		rbx, 64
		sub		rbx, _t2
		mov		rcx, _t2

MainLoop:

		mov		rax, qword ptr[r11+rdi*8]
		shl		rax, cl
		add		qword ptr[r9+rsi*8], rax
		mov		rax, qword ptr[r11+rdi*8]
		xchg	rbx, rcx
		shr		rax, cl
		inc		rdi
		inc		rsi
		mov		qword ptr[r9+rsi*8], rax
		xchg	rbx, rcx
		cmp		rdi, qword ptr[r11]
		jna		MainLoop
		mov		rcx, r9
		CHECKFORZEROS
		mov		r9, rax
		mov		rsi, r9
		mov		rdi, r11
		mov		rcx, qword ptr[r9]
		inc		rcx
		pushf
		cld
rep		movsq
		popf

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	ADivBitShiftLeft64 ENDP
	;;
	;@This is a special version of the BitShiftRight routine used in the divide and mod routines.
	;
	;This is basically the same alg as the BitShiftRight routine except instead of creating a return array it uses the temp array from the divide/mod algs
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	ADivBitShiftRight64 PROC val:PTR qword, shift:qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		rax, rdx
		xor		rdx, rdx
		div		_sixtyfour
		mov		_t1, rax
		mov		_t2, rdx
		cmp		rax, qword ptr[r11]
		jnb		ReturnZero
		mov		rax, qword ptr[r11]
		sub		rax, _t1
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_jfunc, r11
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r11, _jfunc
		mov		r8, rax
		mov		rbx, 1
		cmp		r8, 0
		je		MallocError
		mov		rsi, qword ptr[r11]
		mov		rax, rsi
		sub		rax, _t1
		mov		qword ptr[r8], rax
		mov		rcx, rax
		sub		rsi, rax
		inc		rsi
		mov		rax, rsi
		mul		_eight
		mov		rsi, r11
		add		rsi, rax
		mov		rdi, r8
		add		rdi, 8
		pushf
		cld
rep		movsq
		popf
		cmp		_t2, 0
		je		GoOut
		mov		rbx, 64
		mov		rcx, _t2
		sub		rbx, rcx
		mov		rsi, 1
		mov		rdi, 2

ShiftLoop:

		shr		qword ptr[r8+rsi*8], cl
		xchg	rbx, rcx
		cmp		rdi, qword ptr[r8]
		ja		GoOut
		mov		rax, qword ptr[r8+rdi*8]
		shl		rax, cl
		add		qword ptr[r8+rsi*8], rax
		xchg	rbx, rcx
		inc		rdi
		inc		rsi
		jmp		ShiftLoop

ReturnZero:

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		cmp		r8, 0
		je		MallocError
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		jmp		GoOut

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	ADivBitShiftRight64 ENDP
	;;
	;@This is a special version of the ByteShiftLeft routine used in the divide and mod routines.
	;
	;This is basically the same alg as the ByteShiftLeft routine except instead of creating a return array it uses the temp array from the divide/mod algs
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	ADivByteShiftLeft64 PROC val:PTR qword, shift:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		_t1, rdx
		mov		rdi, qword ptr[r11]
		mov		rsi, rdi
		mov		rax, rdi
		mov		rcx, rax
		add		rax, _t1
		mov		r8, rax
		mul		_eight
		mov		rdi, r11
		add		rdi, rax
		mov		rax, rsi
		mul		_eight
		mov		rsi, r11
		add		rsi, rax
		pushf
		std
rep		movsq
		mov		rcx, _t1
		mov		rax, 0
rep		stosq
		popf
		mov		qword ptr[r11], r8
		mov		rax, r11

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	ADivByteShiftLeft64 ENDP
	;;
	;@This is a special version of the ByteShiftRight routine used in the divide and mod routines.
	;
	;This is basically the same alg as the ByteShiftRight routine except instead of creating a return array it uses the temp array from the divide/mod algs
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	ADivByteShiftRight64 PROC val:PTR qword, shift:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		_t1, rdx
		mov		rax, qword ptr[r11]
		cmp		rax, _t1
		jna		GoOut
		mov		rax, qword ptr[r11]
		sub		rax, _t1
		mov		rcx, rax
		mov		r8, rax
		mov		rdi, r11
		add		rdi, 8
		mov		rax, _t1
		inc		rax
		mul		_eight
		mov		rsi, r11
		add		rsi, rax
		pushf
		cld
rep		movsq
		popf

GoOut:

		mov		qword ptr[r11], r8
		mov		rax, r11

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	ADivByteShiftRight64 ENDP
	;;
	;@brief Divides 2 arrays.
	;
	;This alg starts by making a series of checks, is the divisor equal to 0? then error out.  Is the divisor equal to one? then return _val1, is the divisor equal to _val1, then return 1, is the divisor
	;greater then _val1? then return 0.  If it passes all the tests, then it creates a copy of _val1, this will be the value we keep subtracting from so we dont change the original.  Then it creates a temp
	;variable the same size as _val1.  Then we compute the size of the return value by subtracting the number of bits in the divisor from the number of bits in _val1 and dividing that by 64 (plus one
	;if rdx greater than 0)  This tells us how many elements might be in the result (it could be one less if the most significant byte of _val1 is greater than the most significant byte of the divisor
	;after all of our preparations) , zero out the elements and set element zero to the length.  After that, we check the most significant byte of the divisor to see if it is 64 bits long.  If it is we go straight to the main loop.  If not, we make a copy of the 
	;divisor, then shift all the elements in the divisor by however many bits we need to to make that byte 64 bits long, and we shift _val1 by the same number of bits to the left.  This is to minimize
	;the renormalization problem, with this tactic, the only time renormalization takes place is one time during the first run of the main loop.  Before entering the main loop, we check the most significant
	;bytes of _val1 and the divisor, if _val1 msb is less than the divisors msb we flag it for two variables.  Then we calc zeros for the temp variable.
	;
	;  Once in the main loop, we check the flag for two variables, if it is set then we move the most significant byte of _val1 into rdx, and the next byte into rcx, if not, we zero out rdx and move the most
	;significant byte into rcx.  Then we divide by the most significant byte in the divisor, this is saved in the _retval, then the DivMult routine is called which multiplies the answer from our div to the
	;divisor and stores it in the temp variable.  Then we call the DivByteShiftLeft to shift all the elements to the left so they are lined up for the subtraction, and it zeros out all the low end bytes.
	;next we check to make sure the result from all of this is smaller then _val1, if its not, we decrement the result from the dision by one and redo the above (in practice this only seems to happen once 
	;on the first run if at all).  If it is smaller, we call the DivSub routine which subtracts the temp variable from _val1.  Then we check the most significant byte of _val1 (this is the most significant
	;byte before the subtraction) against the most significant byte of the divisor.  If it is greater, we increment the result of the division and go through all the above again (and again, if this happens
	;at all it only happens on the first run).  If the most significant byte of _val1 is zero we set the two variable flag to 0 and jump to the last section of the loop before jumping to the beggining, if its not zero, we set the two variable flag to 1
	;then drop through to the last section before jumping to the head of the loop to do it over again.  The last section decrements the position in the _retval that we are saving to, if its zero we jump
	;out, if not we check to make sure _val1 length is at least equal to the divisor length, if its not we jump out, otherwise we jump to the head of the main loop and do it again.
	;
	;  Finally, before exiting, we delete all the temp variables and copies we made, send the _retval to the CheckForZeros routine, then exit.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	ADivide64 PROC dtemp:PTR qword, dval2:PTR qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _zero, _one, _one, _zero, _one, _one, _zero

		;start the alg by checking to see if _val1 or _val2 is 0 or 1 and if _one of them are take the appropriate action

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var3, rcx
		mov		_var4, rdx
		mov		rcx, _var3
		mov		rdx, _var4
		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, 0
		mov		_var2, 0
		mov		_t2, 0
		mov		_t3, 0
		mov		r11, rcx
		mov		r12, rdx
		mov		_t1, 1
		cmp		qword ptr[r12], 0
		je		GoOut
		cmp		qword ptr[r12], 1
		ja		CheckV1
		cmp		qword ptr[r12+8], 0
		je		GoOut
		cmp		qword ptr[r12+8], 1
		je		ReturnVal1
		cmp		qword ptr[r11], 2
		ja		GetRetLength
		jb		JustOne
		mov		rdx, qword ptr[r11+16]
		cmp		rdx, qword ptr[r12+8]
		jnb		GetRetLength
		mov		rax, qword ptr[r11+8]
		div		qword ptr[r12+8]
		mov		_jfunc, rax
		mov		rcx, 24
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		mov		rax, _jfunc
		cmp		r8, 0
		je		MallocError
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], rax
		jmp		GoOut

JustOne:

		;if they both only have _one element then just do the division and return the answer

		xor		rdx, rdx
		mov		rax, qword ptr[r11+8]
		div		qword ptr[r12+8]
		mov		_jfunc, rax
		mov		rcx, 24
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		mov		rax, _jfunc
		cmp		r8, 0
		je		MallocError
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], rax
		jmp		GoOut

CheckV1:

		;this is to make sure _val1 gets checked, this part can be hit because of a branch

		cmp		qword ptr[r11], 0
		je		ReturnZero
		cmp		qword ptr[r11], 1
		ja		NextCheck
		cmp		qword ptr[r11+8], 0
		je		ReturnZero

NextCheck:

		;make sure the length of _val1 is equal to or greater than the length of _val2

		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r12]
		jb		ReturnZero
		ja		GetRetLength

CmpLoop:

		;if they are the same length make sure _val1 is greater than _val2, return _one if they are equal

		mov		rax, qword ptr[r11+rdi*8]
		cmp		rax, qword ptr[r12+rdi*8]
		jb		ReturnZero
		ja		GetRetLength
		dec		rdi
		cmp		rdi, 0
		je		ReturnOne
		jmp		CmpLoop

GetRetLength:

		;count how many of the high digits are equal, this is to fix the problem where the division comes out to 1 but should be FFFFFFFFFFFFFFFF, the firstpass variable is used in cunjuction
		;because typically the size off the return variable will have to be reduced by _one on the first correction pass only

		mov		rdi, qword ptr[r11]
		mov		rsi, qword ptr[r12]

SameLoop:

		mov		rax, qword ptr[r11+rdi*8]
		cmp		rax, qword ptr[r12+rsi*8]
		jne		SameOut
		inc		_t3
		dec		rdi
		sub		rsi, 1
		jnz		SameLoop

SameOut:

		;computer the length of the return _val by adding all the _bits in _val1 and subtracting the _bits in _val2 (this part isnt perfect) and dividing by 64

		mov		rax, qword ptr[r11]

FindByteLoop1:

		cmp		qword ptr[r11+rax*8], 0
		ja		FoundByte1
		cmp		rax, 1
		ja		JustDecIt1
		mov		rax, rax

JustDecIt1:

		dec		rax
		jmp		FindByteLoop1

FoundByte1:

		bsr		rcx, qword ptr[r11+rax*8]
		dec		rax
		mul		_sixtyfour
		add		rcx, rax
		mov		rax, qword ptr[r12]

FindByteLoop2:

		cmp		qword ptr[r12+rax*8], 0
		ja		FoundByte2
		cmp		rax, 1
		ja		JustDecIt2
		mov		rax, rax

JustDecIt2:

		dec		rax
		jmp		FindByteLoop2

FoundByte2:

		bsr		rbx, qword ptr[r12+rax*8]
		dec		rax
		mul		_sixtyfour
		add		rbx, rax
		sub		rcx, rbx
		mov		rax, rcx
		inc		rax
		xor		rdx, rdx
		div		_sixtyfour
		cmp		rdx, 0
		je		NoInc
		inc		rax

NoInc:

		;get the memory for the return val

		mov		_jfunc, rax
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		mov		rax, _jfunc
		cmp		r8, 0
		je		MallocError
		mov		qword ptr[r8], rax
		mov		rdi, r8
		add		rdi, 8
		mov		rcx, rax
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf

		;_zero out the return val then get the memory to make a copy of val1(we do this so we dont change the original value when shifting and subtracting)

		mov		rax, qword ptr[r11]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r9, rax
		mov		rsi, r11
		mov		rdi, r9
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf

		;copy val1 into the new array then if the most significant byte in val2 is not 64 _bits get the memory to copy val2

		mov		r11, r9
		mov		_var1, 1
		mov		rax, qword ptr[r11]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r9, rax
		mov		_t2, 1
		mov		r15, qword ptr[r8]
		mov		rdi, qword ptr[r12]
		bsr		rax, qword ptr[r12+rdi*8]
		mov		rcx, 63
		sub		rcx, rax
		cmp		rcx, 0
		je		MainLoop
		mov		_jfunc, rcx
		mov		rax, qword ptr[r12]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		rbx, rax
		mov		rsi, r12
		mov		rdi, rbx
		mov		rcx, qword ptr[r12]
		inc		rcx
		pushf
		cld
rep		movsq
		popf

		;copy val2 into the new array then shift it by however many _bits to make the most significant byte 64 _bits long, then shift val1 the same amount
		
		mov		r12, rbx
		mov		_var2, 1
		mov		rcx, r12
		mov		rdx, _jfunc
		DivBitShiftLeft64
		mov		rcx, r11
		mov		rdx, _jfunc
		DivBitShiftLeft64

MainLoop:

		;this is the main loop, here we figure out if we need to fill just rax or if we need to fill rax, rdx

		xor		rdx, rdx
		mov		rdi, qword ptr[r11]
		mov		rsi, qword ptr[r12]
		mov		rax, qword ptr[r11+rdi*8]
		dec		rdi
		cmp		rdi, 0
		je		KeepThis
		cmp		rax, qword ptr[r12+rsi*8]
		jnb		KeepThis
		mov		rdx, rax
		mov		rax, qword ptr[r11+rdi*8]

KeepThis:

		;do the division and move the value into r14, check if _t3 greater then 0, if it is check if r14==1, if _t1==1 then the size of the return val will have to be adjusted
		;this is d_one in a separate section which is why it passes through here

		div		qword ptr[r12+rsi*8]
		mov		r14, rax
		cmp		_t3, 0
		je		DoItAgain
		cmp		r14, 1
		jne		DoItAgain
		cmp		_t1, 1
		je		DoItAgain
		sub		r14, 2
		dec		_t3

DoItAgain:

		;save the value in the return val then multiply val2 by the value in r14 and store the result in r9
		;then we do a quick check of the lengths of _val1 and _val2, if _val2 is larger then it _zeros the result just saved in the return val and shifts the r9 value to the right _one

		mov		qword ptr[r8+r15*8], r14
		cmp		r14, 0
		je		NextRetByte
		mov		rcx, r12
		mov		rdx, r14
		mov		rax, r9
		DivMult64
		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r9]
		jb		SomethingIsWrong
		mov		rax, r15
		dec		rax
		cmp		rax, 0
		je		DoLastCheck
		mov		rcx, r9
		mov		rdx, rax
		DivByteShiftLeft64

DoLastCheck:

		;here we start checking the values to make sure the value in r9 is less than or equal to val1

		mov		rbx, 0
		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r9]
		ja		DoSub
		jb		SomethingIsWrong
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rsi, r11
		add		rsi, rax
		mov		rdi, r9
		add		rdi, rax
		pushf
		std
repe	cmpsq
		ja		PrepDoSub
		jb		PrepDecAndRedo
		popf
		jmp		DoSub

PrepDecAndRedo:

		popf
		jmp		DecAndRedo

PrepDoSub:

		popf

DoSub:

		;subtract the value in r9 from the value in val1

		mov		rcx, r11
		mov		rdx, r9
		DivSub64

		;this is the final check, it makes sure that r14 was large enough if it wasnt it is supposed to inc r14 and doitagain(but it isnt running that way right now)

		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r12]
		ja		NextRetByte
		jb		GoOut
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rsi, r11
		add		rsi, rax
		mov		rdi, r12
		add		rdi, rax
		pushf
		std
repe	cmpsq
		ja		PrepNextRetByte
		popf
		jmp		GoOut

PrepNextRetByte:

		popf
		jmp		NextRetByte

DecAndRedo:

		;if the value in r14 was too large we dec r14 and doitagain, this is also the area where we adjust the retval size and adjust r14 if we are looking at equal most significant bytes

		cmp		r14, 1
		ja		JustDec
		cmp		_t3, 0
		je		JustDec
		sub		r14, 2
		dec		_t3
		cmp		_t1, 0
		je		DoItAgain
		mov		_t1, 0
		mov		qword ptr[r8+r15*8], 0
		dec		r15
		jmp		DoItAgain

JustDec:

		dec		r14
		jmp		DoItAgain

SomethingIsWrong:

		;something was wrong with how we calculated the length of the return val, so _zero out the return val, store r14 in the next element down, shift r9 to the right _one, and redo the check

		mov		qword ptr[r8+r15*8], 0
		dec		r15
		cmp		r15, 0
		je		GoOut
		mov		qword ptr[r8+r15*8], r14
		mov		rcx, r9
		mov		rdx, 1
		DivByteShiftRight64
		mov		r9, rax
		jmp		DoLastCheck

NextRetByte:

		;everything went ok so on to the next byte

		dec		r15
		cmp		r15, 0
		je		GoOut
		jmp		MainLoop

ReturnZero:

		;return _zero without doing the division

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		jmp		OutOfHere

ReturnOne:

		;return _one without doing the division

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 1
		jmp		OutOfHere

ReturnVal1:

		;return val1 without doing the division

		mov		rax, qword ptr[r11]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		jmp		OutOfHere

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		;do our check for any leading _zeros and dec the element at position 0

		mov		rcx, r8
		CHECKFORZEROS
		mov		r8, rax

OutOfHere:

		cmp		_var1, 0
		je		V1NotCreated
		mov		rcx, r11
		call	Free64

V1NotCreated:

		cmp		_var2, 0
		je		V2NotCreated
		mov		rcx, r12
		call	Free64

V2NotCreated:

		cmp		_t2, 0
		je		TempNotCreated
		mov		rcx, r9
		call	Free64

TempNotCreated:

		;exit the alg

		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _zero, _one, _one, _zero, _one, _one, _zero

		RET

	ADivide64 ENDP
	;;
	;@brief Modular reduction of an array.
	;
	;This alg starts by making a series of checks, is the divisor equal to 0? then error out.  Is the divisor equal to one? then return _val1, is the divisor equal to _val1, then return 1, is the divisor
	;greater then _val1? then return 0.  If it passes all the tests, then it creates a copy of _val1, this will be the value we keep subtracting from so we dont change the original.  Then it creates a temp
	;variable the same size as _val1.  Then we compute the size of the return value by subtracting the number of bits in the divisor from the number of bits in _val1 and dividing that by 64 (plus one
	;if rdx greater than 0)  This tells us how many elements might be in the result (it could be one less if the most significant byte of _val1 is greater than the most significant byte of the divisor
	;after all of our preparations) , zero out the elements and set element zero to the length.  After that, we check the most significant byte of the divisor to see if it is 64 bits long.  If it is we go straight to the main loop.  If not, we make a copy of the 
	;divisor, then shift all the elements in the divisor by however many bits we need to to make that byte 64 bits long, and we shift _val1 by the same number of bits to the left.  This is to minimize
	;the renormalization problem, with this tactic, the only time renormalization takes place is one time during the first run of the main loop.  Before entering the main loop, we check the most significant
	;bytes of _val1 and the divisor, if _val1 msb is less than the divisors msb we flag it for two variables.  Then we calc zeros for the temp variable.
	;
	;  Once in the main loop, we check the flag for two variables, if it is set then we move the most significant byte of _val1 into rdx, and the next byte into rcx, if not, we zero out rdx and move the most
	;significant byte into rcx.  Then we divide by the most significant byte in the divisor, this is saved in the _retval, then the DivMult routine is called which multiplies the answer from our div to the
	;divisor and stores it in the temp variable.  Then we call the DivByteShiftLeft to shift all the elements to the left so they are lined up for the subtraction, and it zeros out all the low end bytes.
	;next we check to make suure the result from all of this is smaller then _val1, if its not, we decrement the result from the dision by one and redo the above (in practice this only seems to happen once 
	;on the first run if at all).  If it is smaller, we call the DivSub routine which subtracts the temp variable from _val1.  Then we check the most significant byte of _val1 (this is the most significant
	;byte before the subtraction) against the most significant byte of the divisor.  If it is greater, we increment the result of the division and go through all the above again (and again, if this happens
	;at all it only happens on the first run).  If the most significant byte of _val1 is zero we set the two variable flag to 0 and jump to the last section of the loop before jumping to the beggining, if its not zero, we set the two variable flag to 1
	;then drop through to the last section before jumping to the head of the loop to do it over again.  The last section decrements the position in the _retval that we are saving to, if its zero we jump
	;out, if not we check to make sure _val1 length is at least equal to the divisor length, if its not we jump out, otherwise we jump to the head of the main loop and do it again.
	;
	;  Finally, before exiting, we delete all the temp variables and copies we made, we shift the _retval to the right the same number of bits we shifted to the left at the beggining, send the _retval to 
	;the CheckForZeros routine, then exit.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AMod64 PROC dtemp:PTR qword, dval2:PTR qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _zero, _one, _one, _zero, _one, _one, _zero

		;start the alg by checking to see if _val1 or _val2 is 0 or 1 and if _one of them are take the appropriate action

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var5, rcx
		mov		_var6, rdx
		mov		rcx, _var5
		mov		rdx, _var6
		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, 0
		mov		_var2, 0
		mov		_t2, 0
		mov		_var4, 0
		mov		_var3, 0
		mov		_t3, 0
		mov		r11, rcx
		mov		r12, rdx
		mov		_t1, 1
		cmp		qword ptr[r12], 0
		je		GoOut
		cmp		qword ptr[r12], 1
		ja		CheckV1
		cmp		qword ptr[r12+8], 0
		je		GoOut
		cmp		qword ptr[r12+8], 1
		je		ReturnZero
		cmp		qword ptr[r11], 2
		ja		GetRetLength
		jb		JustOne
		mov		rdx, qword ptr[r11+16]
		cmp		rdx, qword ptr[r12+8]
		jnb		GetRetLength
		mov		rax, qword ptr[r11+8]
		div		qword ptr[r12+8]
		mov		_jfunc, rdx
		mov		rcx, 24
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		mov		rax, _jfunc
		cmp		r8, 0
		je		MallocError
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], rax
		mov		r11, r8
		jmp		GoOut

JustOne:

		;if they both only have _one element then just do the division and return the answer

		xor		rdx, rdx
		mov		rax, qword ptr[r11+8]
		div		qword ptr[r12+8]
		mov		_jfunc, rdx
		mov		rcx, 24
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		mov		rax, _jfunc
		cmp		r8, 0
		je		MallocError
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], rax
		mov		r11, r8
		jmp		GoOut

CheckV1:

		;this is to make sure _val1 gets checked, this part can be hit because of a branch

		cmp		qword ptr[r11], 0
		je		ReturnZero
		cmp		qword ptr[r11], 1
		ja		NextCheck
		cmp		qword ptr[r11+8], 0
		je		ReturnZero

NextCheck:

		;make sure the length of _val1 is equal to or greater than the length of _val2

		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r12]
		jb		ReturnVal1
		ja		GetRetLength

CmpLoop:

		;if they are the same length make sure _val1 is greater than _val2, return _one if they are equal

		mov		rax, qword ptr[r11+rdi*8]
		cmp		rax, qword ptr[r12+rdi*8]
		jb		ReturnVal1
		ja		GetRetLength
		dec		rdi
		cmp		rdi, 0
		je		ReturnZero
		jmp		CmpLoop

GetRetLength:

		;count how many of the high digits are equal, this is to fix the problem where the division comes out to 1 but should be FFFFFFFFFFFFFFFF, the firstpass variable is used in cunjuction
		;because typically the size off the return variable will have to be reduced by _one on the first correction pass only

		mov		rdi, qword ptr[r11]
		mov		rsi, qword ptr[r12]
		cmp		rdi, 1000
		jb		SameLoop
		mov		rax, rax

SameLoop:

		mov		rax, qword ptr[r11+rdi*8]
		cmp		rax, qword ptr[r12+rsi*8]
		jne		SameOut
		inc		_t3
		dec		rdi
		sub		rsi, 1
		jnz		SameLoop

SameOut:

		;computer the length of the return _val by adding all the _bits in _val1 and subtracting the _bits in _val2 (this part isnt perfect) and dividing by 64

		mov		rax, qword ptr[r11]

FindByteLoop1:

		cmp		qword ptr[r11+rax*8], 0
		ja		FoundByte1
		cmp		rax, 1
		ja		JustDecIt1
		mov		rax, rax

JustDecIt1:

		dec		rax
		jmp		FindByteLoop1

FoundByte1:

		bsr		rcx, qword ptr[r11+rax*8]
		dec		rax
		mul		_sixtyfour
		add		rcx, rax
		mov		rax, qword ptr[r12]

FindByteLoop2:

		cmp		qword ptr[r12+rax*8], 0
		ja		FoundByte2
		cmp		rax, 1
		ja		JustDecIt2
		mov		rax, rax

JustDecIt2:

		dec		rax
		jmp		FindByteLoop2

FoundByte2:

		bsr		rbx, qword ptr[r12+rax*8]
		dec		rax
		mul		_sixtyfour
		add		rbx, rax
		sub		rcx, rbx
		mov		rax, rcx
		inc		rax
		xor		rdx, rdx
		div		_sixtyfour
		cmp		rdx, 0
		je		NoInc
		inc		rax

NoInc:

		;get the memory for the return val

		mov		_jfunc, rax
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		mov		_var4, 1
		mov		rax, _jfunc
		cmp		r8, 0
		je		MallocError

		;_zero out the return val then get the memory to make a copy of val1(we do this so we dont change the original value when shifting and subtracting)

		mov		qword ptr[r8], rax
		mov		rdi, r8
		add		rdi, 8
		mov		rcx, rax
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf
		mov		rax, qword ptr[r11]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r9, rax

		;copy val1 into the new array then if the most significant byte in val2 is not 64 _bits get the memory to copy val2

		mov		rsi, r11
		mov		rdi, r9
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		r11, r9
		mov		_var1, 1
		mov		rax, qword ptr[r11]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r9, rax
		mov		_t2, 1
		mov		r15, qword ptr[r8]
		mov		rdi, qword ptr[r12]
		bsr		rax, qword ptr[r12+rdi*8]
		mov		rcx, 63
		sub		rcx, rax
		cmp		rcx, 0
		je		MainLoop
		mov		_jfunc, rcx
		mov		rax, qword ptr[r12]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		rbx, rax

		;copy val2 into the new array then shift it by however many _bits to make the most significant byte 64 _bits long, then shift val1 the same amount

		mov		rsi, r12
		mov		rdi, rbx
		mov		rcx, qword ptr[r12]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		r12, rbx
		mov		_var2, 1
		mov		rcx, r12
		mov		rdx, _jfunc
		mov		_var3, rdx
		DivBitShiftLeft64
		mov		rcx, r11
		mov		rdx, _jfunc
		DivBitShiftLeft64

MainLoop:

		;this is the main loop, here we figure out if we need to fill just rax or if we need to fill rax, rdx

		xor		rdx, rdx
		mov		rdi, qword ptr[r11]
		mov		rsi, qword ptr[r12]
		mov		rax, qword ptr[r11+rdi*8]
		dec		rdi
		cmp		rdi, 0
		je		KeepThis
		cmp		rax, qword ptr[r12+rsi*8]
		jnb		KeepThis
		mov		rdx, rax
		mov		rax, qword ptr[r11+rdi*8]

KeepThis:

		;do the division and move the value into r14, check if _t3 greater then 0, if it is check if r14==1, if _t1==1 then the size of the return val will have to be adjusted
		;this is d_one in a separate section which is why it passes through here

		div		qword ptr[r12+rsi*8]
		mov		r14, rax
		cmp		_t3, 0
		je		DoItAgain
		cmp		r14, 1
		jne		DoItAgain
		cmp		_t1, 1
		je		DoItAgain
		sub		r14, 2
		dec		_t3

DoItAgain:

		;save the value in the return val then multiply val2 by the value in r14 and store the result in r9
		;then we do a quick check of the lengths of _val1 and _val2, if _val2 is larger then it _zeros the result just saved in the return val and shifts the r9 value to the right _one

		mov		qword ptr[r8+r15*8], r14
		cmp		r14, 0
		je		NextRetByte
		mov		rcx, r12
		mov		rdx, r14
		mov		rax, r9
		DivMult64
		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r9]
		jb		SomethingIsWrong
		mov		rax, r15
		dec		rax
		cmp		rax, 0
		je		DoLastCheck
		mov		rcx, r9
		mov		rdx, rax
		DivByteShiftLeft64

DoLastCheck:

		;here we start checking the values to make sure the value in r9 is less than or equal to val1

		mov		rbx, 0
		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r9]
		ja		DoSub
		jb		SomethingIsWrong
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rsi, r11
		add		rsi, rax
		mov		rdi, r9
		add		rdi, rax
		pushf
		std
repe	cmpsq
		ja		PrepDoSub
		jb		PrepDecAndRedo
		popf
		jmp		ReturnZeroMade

PrepDecAndRedo:

		popf
		jmp		DecAndRedo
		
PrepDoSub:

		popf

DoSub:

		;subtract the value in r9 from the value in val1

		mov		rcx, r11
		mov		rdx, r9
		DivSub64

		;this is the final check, it makes sure that r14 was large enough if it wasnt it is supposed to inc r14 and doitagain(but it isnt running that way right now)

		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r12]
		ja		NextRetByte
		jb		GoOut
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rsi, r11
		add		rsi, rax
		mov		rdi, r12
		add		rdi, rax
		pushf
		std
repe	cmpsq
		ja		PrepNextRetByte
		popf
		jmp		GoOut

PrepNextRetByte:

		popf
		jmp		NextRetByte

DecAndRedo:

		;if the value in r14 was too large we dec r14 and doitagain, this is also the area where we adjust the retval size and adjust r14 if we are looking at equal most significant bytes

		cmp		r14, 1
		ja		JustDec
		cmp		_t3, 0
		je		JustDec
		sub		r14, 2
		dec		_t3
		cmp		_t1, 0
		je		DoItAgain
		mov		_t1, 0
		mov		qword ptr[r8+r15*8], 0
		dec		r15
		jmp		DoItAgain

JustDec:

		dec		r14
		jmp		DoItAgain

SomethingIsWrong:

		;something was wrong with how we calculated the length of the return val, so _zero out the return val, store r14 in the next element down, shift r9 to the right _one, and redo the check

		mov		qword ptr[r8+r15*8], 0
		dec		r15
		cmp		r15, 0
		je		GoOut
		mov		qword ptr[r8+r15*8], r14
		mov		rcx, r9
		mov		rdx, 1
		DivByteShiftRight64
		mov		r9, rax
		jmp		DoLastCheck

NextRetByte:

		;everything went ok so on to the next byte

		dec		r15
		cmp		r15, 0
		je		GoOut
		jmp		MainLoop

ReturnZeroMade:

		mov		qword ptr[r11], 1
		mov		qword ptr[r11+8], 0
		jmp		OutOfHere

ReturnZero:

		;return _zero without doing the division

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		mov		r11, r8
		jmp		OutOfHere

ReturnOne:

		;return _one without doing the division

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 1
		mov		r11, r8
		jmp		OutOfHere

ReturnVal1:

		;return val1 without doing the division

		mov		rax, qword ptr[r11]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		r11, r8
		jmp		OutOfHere

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		;do our check for any leading _zeros and dec the element at position 0

OutOfHere:

		cmp		_var4, 0
		je		RetNotCreated
		mov		rcx, r8
		call	Free64

RetNotCreated:

		cmp		_var2, 0
		je		V2NotCreated
		mov		rcx, r12
		call	Free64

V2NotCreated:

		cmp		_t2, 0
		je		TempNotCreated
		mov		rcx, r9
		call	Free64

TempNotCreated:

		cmp		_var3, 0
		je		DontShiftIt
		mov		rcx, r11
		mov		rdx, _var3
		DivBitShiftRight64
		cmp		_var1, 0
		je		DontShiftIt
		mov		rcx, r11
		mov		_jfunc, rax
		call	Free64
		mov		r11, _jfunc

DontShiftIt:

		mov		rcx, r11
		CHECKFORZEROS
		mov		r11, rax

ExitNow:

		;exit the alg

		mov		rax, r11

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _zero, _one, _one, _zero, _one, _one, _zero

		RET

	AMod64 ENDP
	;;
	;@brief An alg to return both the result of division and the remainder at the same time.
	;
	;This alg starts by making a series of checks, is the divisor equal to 0? then error out.  Is the divisor equal to one? then return _val1, is the divisor equal to _val1, then return 1, is the divisor
	;greater then _val1? then return 0.  If it passes all the tests, then it creates a copy of _val1, this will be the value we keep subtracting from so we dont change the original.  Then it creates a temp
	;variable the same size as _val1.  Then we compute the size of the return value by subtracting the number of bits in the divisor from the number of bits in _val1 and dividing that by 64 (plus one
	;if rdx greater than 0)  This tells us how many elements might be in the result (it could be one less if the most significant byte of _val1 is greater than the most significant byte of the divisor
	;after all of our preparations) , zero out the elements and set element zero to the length.  After that, we check the most significant byte of the divisor to see if it is 64 bits long.  If it is we go straight to the main loop.  If not, we make a copy of the 
	;divisor, then shift all the elements in the divisor by however many bits we need to to make that byte 64 bits long, and we shift _val1 by the same number of bits to the left.  This is to minimize
	;the renormalization problem, with this tactic, the only time renormalization takes place is one time during the first run of the main loop.  Before entering the main loop, we check the most significant
	;bytes of _val1 and the divisor, if _val1 msb is less than the divisors msb we flag it for two variables.  Then we calc zeros for the temp variable.
	;
	;  Once in the main loop, we check the flag for two variables, if it is set then we move the most significant byte of _val1 into rdx, and the next byte into rcx, if not, we zero out rdx and move the most
	;significant byte int rcx.  Then we divide by the most significant byte in the divisor, this is saved in the _retval, then the DivMult routine is called which multiplies the answer from our div to the
	;divisor and stores it in the temp variable.  Then we call the DivByteShiftLeft to shift all the elements to the left so they are lined up for the subtraction, and it zeros out all the low end bytes.
	;next we check to make suure the result from all of this is smaller then _val1, if its not, we decrement the result from the dision by one and redo the above (in practice this only seems to happen once 
	;on the first run if at all).  If it is smaller, we call the DivSub routine which subtracts the temp variable from _val1.  Then we check the most significant byte of _val1 (this is the most significant
	;byte before the subtraction) against the most significant byte of the divisor.  If it is greater, we increment the result of the division and go through all the above again (and again, if this happens
	;at all it only happens on the first run).  If the most significant byte of _val1 is zero we set the two variable flag to 0 and jump to the last section of the loop before jumping to the beggining, if its not zero, we set the two variable flag to 1
	;then drop through to the last section before jumping to the head of the loop to do it over again.  The last section decrements the position in the _retval that we are saving to, if its zero we jump
	;out, if not we check to make sure _val1 length is at least equal to the divisor length, if its not we jump out, otherwise we jump to the head of the main loop and do it again.
	;
	;  Finally, before exiting, we delete all the temp variables and copies we made, we shift the _retval to the right the same number of bits we shifted to the left at the beggining, send the _retval to 
	;the CheckForZeros routine, then exit.
	;@returns an array containing 2 pointers.  The first points to the array from the reult of the division, the second points to the array that represents the remainder.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	ADivRem64 PROC dtemp:PTR qword, dval2:PTR qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _zero, _one, _one, _zero, _one, _one, _zero

		;start the alg by checking to see if _val1 or _val2 is 0 or 1 and if _one of them are take the appropriate action

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var5, rcx
		mov		_var6, rdx
		mov		rcx, _var5
		mov		rdx, _var6
		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, 0
		mov		_var2, 0
		mov		_t2, 0
		mov		_var4, 0
		mov		_var3, 0
		mov		_t3, 0
		mov		r11, rcx
		mov		r12, rdx
		mov		_t1, 1
		cmp		qword ptr[r12], 0
		je		GoOut
		cmp		qword ptr[r12], 1
		ja		CheckV1
		cmp		qword ptr[r12+8], 0
		je		GoOut
		cmp		qword ptr[r12+8], 1
		je		ReturnZero
		cmp		qword ptr[r11], 2
		ja		GetRetLength
		jb		JustOne
		mov		rdx, qword ptr[r11+16]
		cmp		rdx, qword ptr[r12+8]
		jnb		GetRetLength
		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r15, rax
		mov		qword ptr[r15], 1
		mov		rdx, qword ptr[r11+16]
		mov		rax, qword ptr[r11+8]
		div		qword ptr[r12+8]
		mov		qword ptr[r8+8], rax
		mov		qword ptr[r15+8], rdx
		mov		_jfunc, rdx
		mov		rcx, 24
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], rax
		mov		r11, r15
		jmp		GoOut

JustOne:

		;if they both only have _one element then just do the division and return the answer

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r15, rax
		mov		qword ptr[r15], 1
		xor		rdx, rdx
		mov		rax, qword ptr[r11+8]
		div		qword ptr[r12+8]
		mov		qword ptr[r8+8], rax
		mov		qword ptr[r15+8], rdx
		mov		r11, r15
		jmp		GoOut

CheckV1:

		;this is to make sure _val1 gets checked, this part can be hit because of a branch

		cmp		qword ptr[r11], 0
		je		ReturnZero
		cmp		qword ptr[r11], 1
		ja		NextCheck
		cmp		qword ptr[r11+8], 0
		je		ReturnZero

NextCheck:

		;make sure the length of _val1 is equal to or greater than the length of _val2

		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r12]
		jb		ReturnVal1
		ja		GetRetLength

CmpLoop:

		;if they are the same length make sure _val1 is greater than _val2, return _one if they are equal

		mov		rax, qword ptr[r11+rdi*8]
		cmp		rax, qword ptr[r12+rdi*8]
		jb		ReturnVal1
		ja		GetRetLength
		dec		rdi
		cmp		rdi, 0
		je		ReturnZero
		jmp		CmpLoop

GetRetLength:

		;count how many of the high digits are equal, this is to fix the problem where the division comes out to 1 but should be FFFFFFFFFFFFFFFF, the firstpass variable is used in cunjuction
		;because typically the size off the return variable will have to be reduced by _one on the first correction pass only

		mov		rdi, qword ptr[r11]
		mov		rsi, qword ptr[r12]
		cmp		rdi, 1000
		jb		SameLoop
		mov		rax, rax

SameLoop:

		mov		rax, qword ptr[r11+rdi*8]
		cmp		rax, qword ptr[r12+rsi*8]
		jne		SameOut
		inc		_t3
		dec		rdi
		sub		rsi, 1
		jnz		SameLoop

SameOut:

		;computer the length of the return _val by adding all the _bits in _val1 and subtracting the _bits in _val2 (this part isnt perfect) and dividing by 64

		mov		rax, qword ptr[r11]

FindByteLoop1:

		cmp		qword ptr[r11+rax*8], 0
		ja		FoundByte1
		cmp		rax, 1
		ja		JustDecIt1
		mov		rax, rax

JustDecIt1:

		dec		rax
		jmp		FindByteLoop1

FoundByte1:

		bsr		rcx, qword ptr[r11+rax*8]
		dec		rax
		mul		_sixtyfour
		add		rcx, rax
		mov		rax, qword ptr[r12]

FindByteLoop2:

		cmp		qword ptr[r12+rax*8], 0
		ja		FoundByte2
		cmp		rax, 1
		ja		JustDecIt2
		mov		rax, rax

JustDecIt2:

		dec		rax
		jmp		FindByteLoop2

FoundByte2:

		bsr		rbx, qword ptr[r12+rax*8]
		dec		rax
		mul		_sixtyfour
		add		rbx, rax
		sub		rcx, rbx
		mov		rax, rcx
		inc		rax
		xor		rdx, rdx
		div		_sixtyfour
		cmp		rdx, 0
		je		NoInc
		inc		rax

NoInc:

		;get the memory for the return val

		mov		_jfunc, rax
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		mov		_var4, 1
		mov		rax, _jfunc
		cmp		r8, 0
		je		MallocError

		;_zero out the return val then get the memory to make a copy of val1(we do this so we dont change the original value when shifting and subtracting)

		mov		qword ptr[r8], rax
		mov		rdi, r8
		add		rdi, 8
		mov		rcx, rax
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf
		mov		rax, qword ptr[r11]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r9, rax

		;copy val1 into the new array then if the most significant byte in val2 is not 64 _bits get the memory to copy val2

		mov		rsi, r11
		mov		rdi, r9
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		r11, r9
		mov		_var1, 1
		mov		rax, qword ptr[r11]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r9, rax
		mov		_t2, 1
		mov		r15, qword ptr[r8]
		mov		rdi, qword ptr[r12]
		bsr		rax, qword ptr[r12+rdi*8]
		mov		rcx, 63
		sub		rcx, rax
		cmp		rcx, 0
		je		MainLoop
		mov		_jfunc, rcx
		mov		rax, qword ptr[r12]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		rbx, rax

		;copy val2 into the new array then shift it by however many _bits to make the most significant byte 64 _bits long, then shift val1 the same amount

		mov		rsi, r12
		mov		rdi, rbx
		mov		rcx, qword ptr[r12]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		r12, rbx
		mov		_var2, 1
		mov		rcx, r12
		mov		rdx, _jfunc
		mov		_var3, rdx
		DivBitShiftLeft64
		mov		rcx, r11
		mov		rdx, _jfunc
		DivBitShiftLeft64

MainLoop:

		;this is the main loop, here we figure out if we need to fill just rax or if we need to fill rax, rdx

		xor		rdx, rdx
		mov		rdi, qword ptr[r11]
		mov		rsi, qword ptr[r12]
		mov		rax, qword ptr[r11+rdi*8]
		dec		rdi
		cmp		rdi, 0
		je		KeepThis
		cmp		rax, qword ptr[r12+rsi*8]
		jnb		KeepThis
		mov		rdx, rax
		mov		rax, qword ptr[r11+rdi*8]

KeepThis:

		;do the division and move the value into r14, check if _t3 greater then 0, if it is check if r14==1, if _t1==1 then the size of the return val will have to be adjusted
		;this is d_one in a separate section which is why it passes through here

		div		qword ptr[r12+rsi*8]
		mov		r14, rax
		cmp		_t3, 0
		je		DoItAgain
		cmp		r14, 1
		jne		DoItAgain
		cmp		_t1, 1
		je		DoItAgain
		sub		r14, 2
		dec		_t3

DoItAgain:

		;save the value in the return val then multiply val2 by the value in r14 and store the result in r9
		;then we do a quick check of the lengths of _val1 and _val2, if _val2 is larger then it _zeros the result just saved in the return val and shifts the r9 value to the right _one

		mov		qword ptr[r8+r15*8], r14
		cmp		r14, 0
		je		NextRetByte
		mov		rcx, r12
		mov		rdx, r14
		mov		rax, r9
		DivMult64
		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r9]
		jb		SomethingIsWrong
		mov		rax, r15
		dec		rax
		cmp		rax, 0
		je		DoLastCheck
		mov		rcx, r9
		mov		rdx, rax
		DivByteShiftLeft64

DoLastCheck:

		;here we start checking the values to make sure the value in r9 is less than or equal to val1

		mov		rbx, 0
		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r9]
		ja		DoSub
		jb		SomethingIsWrong
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rsi, r11
		add		rsi, rax
		mov		rdi, r9
		add		rdi, rax
		pushf
		std
repe	cmpsq
		ja		PrepDoSub
		jb		PrepDecAndRedo
		popf
		jmp		ReturnZeroMade

PrepDecAndRedo:

		popf
		jmp		DecAndRedo
		
PrepDoSub:

		popf

DoSub:

		;subtract the value in r9 from the value in val1

		mov		rcx, r11
		mov		rdx, r9
		DivSub64

		;this is the final check, it makes sure that r14 was large enough if it wasnt it is supposed to inc r14 and doitagain(but it isnt running that way right now)

		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r12]
		ja		NextRetByte
		jb		GoOut
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rsi, r11
		add		rsi, rax
		mov		rdi, r12
		add		rdi, rax
		pushf
		std
repe	cmpsq
		ja		PrepNextRetByte
		popf
		jmp		GoOut

PrepNextRetByte:

		popf
		jmp		NextRetByte

DecAndRedo:

		;if the value in r14 was too large we dec r14 and doitagain, this is also the area where we adjust the retval size and adjust r14 if we are looking at equal most significant bytes

		cmp		r14, 1
		ja		JustDec
		cmp		_t3, 0
		je		JustDec
		sub		r14, 2
		dec		_t3
		cmp		_t1, 0
		je		DoItAgain
		mov		_t1, 0
		mov		qword ptr[r8+r15*8], 0
		dec		r15
		jmp		DoItAgain

JustDec:

		dec		r14
		jmp		DoItAgain

SomethingIsWrong:

		;something was wrong with how we calculated the length of the return val, so _zero out the return val, store r14 in the next element down, shift r9 to the right _one, and redo the check

		mov		qword ptr[r8+r15*8], 0
		dec		r15
		cmp		r15, 0
		je		GoOut
		mov		qword ptr[r8+r15*8], r14
		mov		rcx, r9
		mov		rdx, 1
		DivByteShiftRight64
		mov		r9, rax
		jmp		DoLastCheck

NextRetByte:

		;everything went ok so on to the next byte

		dec		r15
		cmp		r15, 0
		je		GoOut
		jmp		MainLoop

ReturnZeroMade:

		mov		qword ptr[r11], 1
		mov		qword ptr[r11+8], 0
		jmp		OutOfHere

ReturnZero:

		;return _zero without doing the division

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		mov		rax, qword ptr[r11]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r15, rax
		mov		rdi, rax
		mov		rsi, r11
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		r11, r15
		jmp		GoOut

ReturnOne:

		;return _one without doing the division

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 1
		mov		rax, qword ptr[r11]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r15, rax
		mov		rdi, rax
		mov		rsi, r11
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		r11, r15
		jmp		OutOfHere

ReturnVal1:

		;return val1 without doing the division

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r15, rax
		mov		qword ptr[r15], 1
		mov		qword ptr[r15+8], 0
		mov		rax, qword ptr[r11]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		r11, r15
		jmp		OutOfHere

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		;do our check for any leading _zeros and dec the element at position 0

OutOfHere:

		cmp		_var2, 0
		je		V2NotCreated
		mov		rcx, r12
		call	Free64

V2NotCreated:

		cmp		_t2, 0
		je		TempNotCreated
		mov		rcx, r9
		call	Free64

TempNotCreated:

		cmp		_var3, 0
		je		DontShiftIt
		mov		rcx, r11
		mov		rdx, _var3
		DivBitShiftRight64
		cmp		_var1, 0
		je		DontShiftIt
		mov		rcx, r11
		mov		_jfunc, rax
		call	Free64
		mov		r11, _jfunc

DontShiftIt:

		mov		rcx, r11
		CHECKFORZEROS
		mov		r11, rax
		mov		rcx, r8
		CHECKFORZEROS
		mov		r8, rax
		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		qword ptr[rax], r8
		mov		qword ptr[rax+8], r11

ExitNow:

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _zero, _one, _one, _zero, _one, _one, _zero

		RET

	ADivRem64 ENDP
	;;
	;@This is a special version of the Multiply routine used in the Pow routine.
	;
	;This is basically the same alg as the Multiply routine except instead of creating a return array it uses the temp array from the Pow alg so that we dont keep creating and destroying objects.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	APowMultiply64 PROC mval1:PTR qword, mval2:PTR qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, qword ptr[r11]		
		add		rax, qword ptr[r8]	
		inc		rax	
		mov		qword ptr[r10], rax		
		inc		rax	
		mov		rcx, rax
		mov		rdi, r10
		add		rdi, 8
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf
		mov		rdi, 1
		mov		rsi, 1
		mov		rcx, 1
		mov		rdx, 0

MultLoop:

		mov		rbx, rdx
		mov		rax, qword ptr[r11+rdi*8]
		mul		qword ptr[r8+rsi*8]
		add		qword ptr[r10+rcx*8], rax
		adc		rdx, 0
		add		qword ptr[r10+rcx*8], rbx
		adc		rdx, 0
		inc		rcx
		inc		rdi
		cmp		rdi, qword ptr[r11]
		jna		MultLoop
		mov		qword ptr[r10+rcx*8], rdx
		mov		rdx, 0
		mov		rdi, 1
		inc		rsi
		mov		rcx, rsi
		cmp		rsi, qword ptr[r8]
		jna		MultLoop
		jmp		GoOut

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rax, 0
		RET

GoOut:

		mov		rdi, qword ptr[r10]

CheckForZerosLoop:

		cmp		qword ptr[r10+rdi*8], 0
		ja		OutOfHere
		cmp		rdi, 1
		jna		OutOfHere
		dec		rdi
		mov		qword ptr[r10], rdi
		jmp		CheckForZerosLoop

OutOfHere:

		mov		rax, r10
		mov		r10, r8
		mov		r8, rax

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	APowMultiply64 ENDP
	;;
	;@This is a special version of the Square routine used in the Pow routine.
	;
	;This is basically the same alg as the Square routine except instead of creating a return array it uses the temp array from the Pow alg so that we dont keep creating and destroying objects.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	APowSquare64 PROC val1:PTR qword

		;thiis is a modified version of the multiply alg

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		rax, qword ptr[r8]
		add		rax, rax
		inc		rax
		mov		qword ptr[r10], rax
		mov		rdi, qword ptr[r8]
		cmp		rdi, 1
		ja		Cont
		mov		rax, qword ptr[r8+8]
		cmp		rax, 0
		je		ReturnZero
		cmp		rax, 1
		je		ReturnOne
		mul		rax
		mov		qword ptr[r10+8], rax
		mov		qword ptr[r10], 1
		cmp		rdx, 0
		je		GoOut
		mov		qword ptr[r10+16], rdx
		mov		qword ptr[r10], 2
		jmp		GoOut

Cont:

		mov		rcx, qword ptr[r10]
		mov		rdi, r10
		add		rdi, 8
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf
		mov		rdi, 1
		mov		rsi, 1
		mov		rcx, 1
		mov		rdx, 0

MultLoop:

		mov		rbx, rdx
		mov		rax, qword ptr[r8+rdi*8]
		mul		qword ptr[r8+rsi*8]
		add		rax, rbx
		adc		rdx, 0
		add		qword ptr[r10+rcx*8], rax
		adc		rdx, 0
		inc		rdi
		inc		rcx
		cmp		rdi, qword ptr[r8]
		jna		MultLoop
		mov		qword ptr[r10+rcx*8], rdx
		mov		rdx, 0
		mov		rdi, 1
		inc		rsi
		mov		rcx, rsi
		cmp		rsi, qword ptr[r8]
		jna		MultLoop
		jmp		GoOut

ReturnZero:

		mov		qword ptr[r10], 1
		mov		qword ptr[r10+8], 0
		jmp		OutOfHere

ReturnOne:

		mov		qword ptr[r10], 1
		mov		qword ptr[r10+8], 1
		jmp		OutOfHere

GoOut:

		mov		rdi, qword ptr[r10]

CheckZerosLoop:

		mov		rax, qword ptr[r10+rdi*8]
		cmp		rax, 0
		ja		OutOfHere
		cmp		rdi, 1
		jna		OutOfHere
		dec		rdi
		mov		qword ptr[r10], rdi
		jmp		CheckZerosLoop

OutOfHere:

		mov		rax, r10
		mov		r10, r8
		mov		r8, rax

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	APowSquare64 ENDP

	;;
	;@brief Exponentiation of an array.
	;
	;  This alg starts by doing a series of checks, is the exponent equal to zero? then return one.  Is the exponent equal to one? then return the base.  Is the base equal to zero?  then return zero.
	;Is the base equal to one? then return one.  Afterwards it is the standard repreated square and multiply routine.
	;@returns an array.
	;@param rcx - The Base array.
	;@param rdx - The Exponent array.
	;;
	APow643 PROC _base:PTR qword, _exp:PTR qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _one, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r10, 0
		mov		r11, rcx
		mov		r12, rdx
		cmp		qword ptr[r12], 0
		je		ReturnOne
		cmp		qword ptr[r12], 1
		ja		NextCheck
		cmp		qword ptr[r12+8], 0
		je		ReturnOne
		cmp		qword ptr[r12+8], 1
		je		ReturnBase

NextCheck:

		cmp		qword ptr[r11], 0
		je		ReturnZero
		cmp		qword ptr[r11], 1
		ja		MoveOn
		cmp		qword ptr[r11+8], 0
		je		ReturnZero
		cmp		qword ptr[r11+8], 1
		je		ReturnOne

MoveOn:

		lea		r9, _bits
		mov		rdi, qword ptr[r12]
		mov		_var1, rdi
		mov		rax, qword ptr[r11]
		bsr		rsi, qword ptr[r12+rdi*8]
		mov		_var2, rsi
		add		rax, 10

AddLoop:

		add		rax, rax
		inc		rax
		mov		rcx, qword ptr[r12+rdi*8]
		mov		rdx, qword ptr[r9+rsi*8]
		and		rcx, qword ptr[r9+rsi*8]
		jz		NoAdd
		add		rax, qword ptr[r11]
		inc		rax

NoAdd:

		cmp		rsi, 0
		je		NextByteAdd
		dec		rsi
		jmp		AddLoop

NextByteAdd:

		cmp		rdi, 1
		jna		AddOut
		dec		rdi
		mov		rsi, 63
		jmp		AddLoop

AddOut:

		mul		_eight
		mov		_t1, rax
		mov		rcx, rax
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		rdi, rax
		mov		rsi, r11
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rcx, _t1
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		r10, rax
		cmp		_var2, 0
		ja		MoveOn2
		cmp		_var1, 1
		jna		GoOut
		dec		_var1
		mov		_var2, 64

MoveOn2:

		dec		_var2

MainLoop:

		mov		rcx, r8
		call	APowSquare64
		mov		rdi, _var1
		mov		rcx, qword ptr[r12+rdi*8]
		mov		rsi, _var2
		and		rcx, qword ptr[r9+rsi*8]
		jz		NoMult
		mov		rcx, r11
		mov		rdx, r8
		call	APowMultiply64

NoMult:

		cmp		_var2, 0
		je		NextByte
		dec		_var2
		jmp		MainLoop

NextByte:

		cmp		_var1, 1
		jna		GoOut
		dec		_var1
		mov		_var2, 63
		jmp		MainLoop

GoOut:

		cmp		r10, 0
		je		NotCreated
		mov		rcx, r10
		call	Free64

NotCreated:

		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _one, _zero, _zero

		ret

ReturnZero:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _one, _zero, _zero

		ret

ReturnOne:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _one, _zero, _zero

		ret

ReturnBase:

		mov		rax, qword ptr[r11]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		rdi, rax
		mov		rsi, r11
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rax, r8

MallocError:

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _one, _zero, _zero

		ret

	APow643 endp

	APow64 PROC base:PTR qword, exp:PTR qword

		;saves the registers we use and adjusts the stack pointer

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _one, _zero, _zero

		;prep by saving the base, exponent, and decimal place, moving rcx into r11, rdx into r12, clearing r8 and r10, then begginging our checks

		mov		_v1, rcx					;save the base
		mov		_v2, rdx					;save the exponent
		mov		_var3, rax					;saves the decimal point (this will be kept as the original decimal place variable)
		mov		_var4, rax					;save the decimal point again (this will keep the running total for where the decimal point is)
		mov		r11, rcx					;move the base into r11
		mov		r12, rdx					;move the exponent into r12
		mov		r8, 0						;clear r8
		mov		r10, 0						;clear r10
		cmp		qword ptr[r12], 1			;check if the exponents array has only one element
		ja		NextCheck					;if its more than one jump to the next check because it has to be greater than one
		cmp		qword ptr[r12], 0			;check if the element count is 0 for the exponent
		je		ReturnOne					;technically this should be an error since no array should have an element count of 0, but we treat as if the exponent is equal to 0
		cmp		qword ptr[r12+8], 0			;checck the first value element for 0
		je		ReturnOne					;jump to return one if it does
		cmp		qword ptr[r12+8], 1			;check if the first vaalue element equals 1
		je		ReturnVal1					;jump to returnval1 if it does (this will return a copy of the base)

NextCheck:

		;here we check the base for the same values we checked for the exponent

		cmp		qword ptr[r11], 0			;check if the element count is 0 (again, this should be an error but we treat it as a zero)
		je		ReturnZero					;jump to returnzero if it does
		cmp		qword ptr[r11], 1			;check if the element count is one
		ja		MoveOn						;cancel the rest of the checks if it is above
		cmp		qword ptr[r11+8], 0			;compare the first value element for 0
		je		ReturnZero					;jump to returnzero if it equals
		cmp		qword ptr[r11+8], 1			;compare the first value element for 1
		je		ReturnOne					;jump to return one if it does

MoveOn:

		;first thing to do after the checks is to count how many elements for the answer and temp arrays we need.  I have special versions of the multiply and square routines that do not create
		;a return array (they use the temp array that we create).  Then is a lot faster then creating and destroying the arrays with each pass of the alg.  We overestimate the element count
		;by taking the number of elements in the base to start the running total at, each iteration we double the running total and add one, if the current bit in the exponent is 1
		;we add the base element count to the running total and add one.  In the actual alg that extra byte will not allways be used.  To start we move the exponent element count into rdi and save it in _var1
		;then we count the number of bits in the high byte of the exponent, load the bits array into r9 (this contains the powers of 2 from a power of 0 to a power of 63), move the element count of the
		;base into rax, increase that count a little to give ourselves some room, make sure the bit count in the high byte of the exponent was not 0 (if its not we jump to decandgo if it is we
		;decrement the index into the exponent, make sure the index is not 0, if it is the count is done, if not store 64 in rsiand move on)

		mov		rdi, qword ptr[r12]			;move the element count of the exponent into rdi
		mov		_var1, rdi					;save the element count in _var1
		bsr		rsi, qword ptr[r12+rdi*8]	;count the number of bits in the high byte of the exponent
		mov		_var2, rsi					;save the bit count in _var2
		lea		r9, _bits					;load the address to the bits array into r9 (the bits array have the powers of two)
		mov		rax, qword ptr[r11]			;move the element count of the base into rax
		add		rax, 10						;increase the element count some to give ourselves some room
		cmp		rsi, 0						;compare rsi to 0
		ja		DecAndGo					;if its above jump to decandgo
		cmp		rdi, 1						;if rsi is 0 compare rdi to 1
		je		CountDone					;if rsi equals 1 then the count is done
		mov		rsi, 64						;otherwise store 64 in rsi
		dec		rdi							;decrement the index into the exponent
		mov		_var1, rdi					;save the index in _var2

DecAndGo:

		dec		rsi							;we need to decrement the index into the bits array by 1 because the most significant bit is allready counted
		mov		_var2, rsi					;store the index into the bits array in _var2

CountLoop:

		add		rax, rax					;double rax
		inc		rax							;add one to rax
		mov		rcx, qword ptr[r12+rdi*8]	;move the byte of the exponent we are working on into rcx
		and		rcx, qword ptr[r9+rsi*8]	;and the byte will the value stored in the bits array pointed to by rsi
		jz		NoMult1						;if the result was 0 then dont add the base element count
		add		rax, qword ptr[r11]			;add the element count of the base to rax
		inc		rax							;add one to rax

NoMult1:

		cmp		rsi, 0						;compare the index into the bits array to 0
		ja		NotZero						;if its above 0 jump to notzero
		cmp		rdi, 1						;otherwise to move to the next byte in the exponent, compare the index into the exponent to 1
		jna		CountDone					;if its not above the count is done
		mov		rsi, 63						;otherwise set the index into the bits array to 63
		dec		rdi							;decrement the index into the exponent
		jmp		CountLoop					;jump to the head of the count loop

NotZero:

		dec		rsi							;the check above meant we still have bits in the current byte to go through so decrement the index to bits
		jmp		CountLoop					;jump to the head of the count loop

CountDone:

		;now that we know how many elements we need for the arrays, we multiply the count by 8 (8 bytes in 64 bits), then make 2 calls to getmem and copy the base to one of them.

		mov		_t1, rax					;move the result of the count into _t1
		mul		_eight						;multiply the result by 8 for 64 bits
		mov		rcx, rax					;move the final result into rcx
		call	GetMem64					;call our getmem routine
		cmp		rax, 0						;make sure what was returned from the routine is valid
		je		MallocError					;if its not jump to mallocerror
		mov		_retval, rax				;move the pointer into _retval
		mov		rcx, qword ptr[r11]			;move the element count of the base into rcx
		inc		rcx							;increment the count by one for the element count at position 0
		mov		rdi, rax					;move the pointer we just got into rdi
		mov		rsi, _v1					;move the pointer for the base into rsi
		pushf								;save the flags register
		cld									;clear the direction flag
rep		movsq								;copy all the elements from the base into the array we just got
		popf								;restore the flags register
		mov		rax, _t1					;move the size of our return array into rax
		mul		_eight						;multiply it by 8 for the 64 bits
		mov		rcx, rax					;move the result into rcx
		call	GetMem64					;call our getmem routine
		cmp		rax, 0						;make sure we got a valid pointer back
		je		MallocError					;if not jump to mallocerror
		mov		r10, rax					;move the pointer into the r10 register
		mov		_t9, r10					;save the pointer in _t9
		mov		r8, _retval					;move _retval into r8

MainLoop:

		;this is just the standard repeated square and multiply alg.  On each iteration we will square the _retval (kept in r8), the special square alg will take r8 and store the result
		;in the array into r10, next we check the current bit of the current byte in the exponent for 1, if it is 1 then we multiply the result from the squaring by the base.

		mov		rax, _var4					;move the running total for the decimal point into rax
		add		_var4, rax					;add rax back into _var4 to double it
		mov		rdi, r10					;move the pointer to the temp array into rdi
		mov		rcx, qword ptr[r8]			;move the element count in _retval into rcx
		add		rcx, rcx					;add rcx to itself to double it
		inc		rcx							;make room for the element count
		inc		rcx							;give it one extra byte in case there is still a carry to do after the final multiply
		mov		qword ptr[r10], rcx			;rcx has our (possible) new element count for the result of the squaring, so store it at position 0 of the temp array
		mov		rax, 0						;move 0 into rax
		add		rdi, 8						;add 8 to rdi so that it points to the element at position one of the temp array
		pushf								;save the flags register
		cld									;clear the direction flag
rep		stosq								;store 0 in all the value elements
		popf								;restore the flags register
		mov		rdi, 1						;set rdi to 1
		mov		rsi, 1						;set rsi to 1
		mov		rcx, 1						;set rcx to 1
		mov		rdx, 0						;clear rdx

SqLoop:

		mov		rbx, rdx					;rdx holds if there was a carry from the last iteration, save it inrbx
		mov		rax, qword ptr[r8+rdi*8]	;move the current byte of the result array into rax
		mul		qword ptr[r8+rsi*8]			;multiply by the other current byte of the result array
		add		qword ptr[r10+rcx*8], rax	;add the result to the current position of the temp array
		adc		rdx, 0						;if there was a carry from the add, this will add one to the carry from the multiply
		add		qword ptr[r10+rcx*8], rbx	;add the carry from the last iteration to the current position in the temp array
		adc		rdx, 0						;if there was a carry from the add, this will add one to the carry from the multiply (only one at most of these adc instructions will inc rdx)
		inc		rcx							;inc our index into the temp array
		inc		rdi							;inc our first position in the _retval
		cmp		rdi, qword ptr[r8]			;compare rdi to the element count
		jna		SqLoop						;if its not larger than jump to the head of the sqloop
		mov		qword ptr[r10+rcx*8], rdx	;the first position reached the end of retval so store the carry from the last iteration in the next element of the temp array (we add it here just in case the second position of the retval has been reached)
		mov		rdx, 0						;zero out the return so we dont add it again
		mov		rdi, 1						;point the first index into retval to the first element
		inc		rsi							;inc the second index into retval
		mov		rcx, rsi					;point the index into the temp array to the same position as the second index into the retval
		cmp		rsi, qword ptr[r8]			;make sure the second index is not past the last element
		jna		SqLoop						;if its not jump to the head of the square loop

SqDone:

		mov		rdi, qword ptr[r10]			;move the element count of the temp array into rdi

ZCheck1:

		cmp		qword ptr[r10+rdi*8], 0		;compare the current element we are pointing at to 0
		ja		ZCheck1Done					;if its above jump out
		cmp		rdi, 1						;otherwise compare the index into the array to one
		jna		ZCheck1Done					;if its not above jump out
		dec		rdi							;otherwise decrement it by one
		mov		qword ptr[r10], rdi			;and store it back in position 0 of the temp array
		jmp		ZCheck1						;jump to the head of the zero check loop

ZCheck1Done:

		mov		rax, r10					;we switch the arrays so that the current result is allways in r8 mmove r10 into rax
		mov		r10, r8						;move r8 into r10
		mov		r8, rax						;move rax into r8
		mov		rdi, _var1					;move the index into the exponent into rdi
		mov		rsi, _var2					;move the index into the bits array into rsi
		mov		rax, qword ptr[r12+rdi*8]	;move the current byte of the exponent into rax
		and		rax, qword ptr[r9+rsi*8]	;and rax with the value pointed to in bits
		jz		NoMult2						;if the result was 0 then jump past the multipliaction loop
		mov		rcx, qword ptr[r8]			;otherwise move the element count of retval into rcx
		add		rcx, qword ptr[r11]			;add the element count of the base to rcx
		inc		rcx							;inc the count for the element count to store in the temp array
		inc		rcx							;inc the element count in case the result of the multiply goes over
		mov		qword ptr[r10], rcx			;store the calcd element count at position 0 of the temp array
		mov		rdi, r10					;move the pointer of the temp array into rdi
		add		rdi, 8						;add 8 to rdi so that we point to the element at position 1
		mov		rax, 0						;move 0 into rax
		pushf								;save the flags register
		cld									;clear the direction flag
rep		stosq								;clear all the elements in the temp array
		popf								;restore the flags register
		mov		rsi, 1						;point to the first element of the retval
		mov		rdi, 1						;point to the first element of the base
		mov		rcx, 1						;point to the first element of the temp array
		mov		rax, _var3					;move the original decimal point into rax
		add		_var4, rax					;add it to the running total
		mov		rdx, 0						;clear rdx

MultLoop:


		mov		rbx, rdx					;again rdx holds the carry from the last iteration so save it in rbx
		mov		rax, qword ptr[r8+rsi*8]	;move the current element of retval into rax
		mul		qword ptr[r11+rdi*8]		;multiply by the current element in the base
		add		qword ptr[r10+rcx*8], rax	;add the result to the current element of the temp array
		adc		rdx, 0						;if the last add resulted in a carry this will add it to the carry from the multiplication
		add		qword ptr[r10+rcx*8], rbx	;add the carry from the last iteration to the current element in the temp array
		adc		rdx, 0						;if the last add resulted in a carry than add it to the carry from the multiply (only one of the adc instructions will inc rdx)
		inc		rcx							;point to the next element in the temp array
		inc		rdi							;point to the next element in the base
		cmp		rdi, qword ptr[r11]			;make sure we are not past the last element in the base
		jna		MultLoop					;if we are good then jump to the head of the multiply loop
		mov		qword ptr[r10+rcx*8], rdx	;otherwise move the carry from this iteration to the new current position in the temp array
		mov		rdx, 0						;clear rdx
		mov		rdi, 1						;point the index into the base at the first element
		inc		rsi							;inc the index into the retval
		mov		rcx, rsi					;point the index into the temp array to the same position as the index into retval
		cmp		rsi, qword ptr[r8]			;make sure we are not past the last element of retval
		jna		MultLoop					;if not jump to the head of the mult loop

MultDone:

		mov		rcx, r10					;move the temp array into rcx
		CHECKFORZEROS						;check for zeros
		mov		r10, r8						;switch the two arrays, move the current retval into r10
		mov		r8, rax						;move the old temp array into retval

NoMult2:

		cmp		_var2, 0					;check if the index into the bits array is 0
		ja		NotZ						;if its not jump to the notz
		cmp		_var1, 1					;otherwise make sure we are still above the least siginificant byte of the exponent
		jna		GoOut						;if not then go out
		dec		_var1						;otherwise dec the index into the exponent
		mov		_var2, 63					;and move 63 into the index into bits
		jmp		MainLoop					;jump to the head of the main loop

NotZ:

		dec		_var2						;we still have bits to go sso decrement the index into the bits array
		jmp		MainLoop					;jump to the head of the main loop

ReturnVal1:

		;the exponent must be equal to one so we return a copy of the base

		mov		rax, qword ptr[r11]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		r11, _v1
		mov		rcx, qword ptr[r11]
		inc		rcx
		mov		rsi, r11
		mov		rdi, r8
		pushf
		cld
rep		movsq
		popf
		mov		rax, qword ptr[r11]
		mov		qword ptr[r8], rax
		jmp		GoOut

ReturnZero:

		;the base must of been equal to 0 (and the exponent was greater than 0) so we return 0

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		mov		_var4, 0
		jmp		GoOut

ReturnOne:

		;either the base was equal to 1 or the exponent was equal to 0

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 1
		mov		_var4, 0
		jmp		GoOut

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		;we are all done so we destroy the temp array if it was created then exit

		cmp		r10, 0
		je		NotC
		mov		rcx, r10
		call	Free64

NotC:

		mov		rcx, _var4
		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _one, _zero, _zero

		RET
	
	APow64	ENDP
	;;
	;@This is an old version of the Pow alg, not used anymore.
	;
	;@returns an array.
	;@param rcx - The base array.
	;@param rdx - The exponent array.
	;;
	APow642 PROC base:PTR qword, exp:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _one, _one, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_t1, rax
		mov		_t2, rax
		mov		_t3, 0
		mov		r11, rcx
		mov		r12, rdx
		cmp		qword ptr[r12], 0
		je		ReturnOne
		cmp		qword ptr[r12], 1
		ja		NextCheck
		cmp		qword ptr[r12+8], 0
		je		ReturnOne
		cmp		qword ptr[r12+8], 1
		je		ReturnV1

NextCheck:

		cmp		qword ptr[r11], 0
		je		ReturnZero
		cmp		qword ptr[r11], 1
		ja		GoOn
		cmp		qword ptr[r11+8], 0
		je		ReturnZero
		cmp		qword ptr[r11+8], 1
		je		ReturnOne

GoOn:

		mov		rcx, qword ptr[r11]
		add		rcx, rcx
		mov		r14, qword ptr[r12]
		bsr		r15, qword ptr[r12+r14*8]
		lea		r9, _bits
		cmp		r15, 0
		je		CheckNextByte
		dec		r15

CheckMainLoop:

		add		rcx, rcx
		inc		rcx
		mov		rax, qword ptr[r9+r15*8]
		test	qword ptr[r12+r14*8], rax
		jz		CheckPassMult
		add		rcx, qword ptr[r11]
		inc		rcx

CheckPassMult:

		cmp		r15, 0
		je		CheckNextByte
		dec		r15
		jmp		CheckMainLoop

CheckNextByte:

		cmp		r14, 1
		jna		CheckGoOut
		mov		r15, 63
		dec		r14
		jmp		CheckMainLoop

CheckGoOut:

		mov		rax, rcx
		add		rax, rax
		add		rax, rax
		mul		_eight
		mov		rcx, rax
		mov		_jfunc, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		cmp		r8, 0
		je		MallocError
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rcx, _jfunc
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r10, rax
		cmp		r10, 0
		je		MallocError
		mov		_t3, 1

		mov		r14, qword ptr[r12]
		bsr		r15, qword ptr[r12+r14*8]
		lea		r9, _bits
		cmp		r15, 0
		je		NextByte
		dec		r15

MainLoop:

		mov		rax, _t2
		add		_t2, rax
		mov		rcx, r8
		PowSquare64
		mov		rax, qword ptr[r9+r15*8]
		test	qword ptr[r12+r14*8], rax
		jz		PassMult
		mov		rax, _t1
		add		_t2, rax
		mov		rcx, r8
		mov		rdx, r11
		PowMultiply64

PassMult:

		cmp		r15, 0
		je		NextByte
		dec		r15
		jmp		MainLoop

NextByte:

		cmp		r14, 1
		jna		GoOut
		mov		r15, 63
		dec		r14
		jmp		MainLoop

ReturnV1:

		mov		rax, qword ptr[r11]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		r11, _v1
		mov		rcx, qword ptr[r11]
		inc		rcx
		mov		rsi, r11
		mov		rdi, r8
		pushf
		cld
rep		movsq
		popf
		mov		rax, qword ptr[r11]
		mov		qword ptr[r8], rax
		jmp		GoOut

ReturnZero:

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		mov		_t2, 0
		jmp		GoOut

ReturnOne:

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 1
		mov		_t2, 0
		jmp		GoOut

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		cmp		_t3, 0
		je		NotCreated
		mov		rcx, r10
		call	Free64

NotCreated:

		mov		rcx, _t2
		mov		rax, r8

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _one, _one, _zero

		RET
	
	APow642	ENDP
	;;
	;@This is a special version of the Multiply routine used in the modPow routine.
	;
	;This is basically the same alg as the Multiply routine except instead of creating a return array it uses the temp array from the modpow alg so that we dont keep creating and destroying objects.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AMPDivMult64 PROC _val1:PTR qword, _val2:qword

		;this is basically the same as the multiply alg except val2 is a single qword value mainly for the divide alg

		ENTERFUNC _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rdx
		mov		rdi, 1
		mov		rdi, 1
		mov		rdx, 0
		mov		rax, qword ptr[r12]
		mov		qword ptr[r9], rax

MainLoop:

		mov		rbx, rdx
		mov		rax, qword ptr[r12+rdi*8]
		mul		r11
		add		rax, rbx
		adc		rdx, 0
		mov		qword ptr[r9+rdi*8], rax
		inc		rdi
		cmp		rdi, qword ptr[r12]
		jna		MainLoop
		cmp		rdx, 0
		je		GoOut
		mov		qword ptr[r9+rdi*8], rdx
		mov		qword ptr[r9], rdi
		jmp		GoOut

GoOut:

		EXITFUNC _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	AMPDivMult64 ENDP
	;;
	;@This is a special version of the Subtract routine used in the modPow routine.
	;
	;This is basically the same alg as the Subtract routine except instead of creating a return array it uses the temp array from the modpow alg so that we dont keep creating and destroying objects.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AMPDivSub64 PROC sval1:PTR qword, sval2:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx

		;the managed code is supposed to have made sure that _val1>=_val2 so we know the return value will be no bigger than _val1

		mov		rdi, 1
		xor		rdx, rdx

SubLoop:

		;here is the main sub loop storing any borrows from the previous loop in rdx

		mov		rcx, rdx
		mov		rax, qword ptr[r9+rdi*8]
		sub		qword ptr[r8+rdi*8], rax
		cmovc	rdx, _one
		cmovnc	rdx, _zero
		sub		qword ptr[r8+rdi*8], rcx
		cmovc	rdx, _one
		inc		rdi
		cmp		rdi, qword ptr[r9]
		jna		SubLoop

SubLoop2:

		;if the length of _val1 is greater than the length of _val2 then the alg continues here basically copying the elements in _val1 with respect to borrows

		cmp		rdi, qword ptr[r8]
		ja		CheckForZs
		sub		qword ptr[r8+rdi*8], rdx
		cmovc	rdx, _one
		cmovnc	rdx, _zero
		inc		rdi
		jmp		SubLoop2

CheckForZs:

		mov		rcx, r8
		CHECKFORZEROS
		mov		r8, rax
		jmp		GoOut

GoOut:

		mov		rax, r8

		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
	
	AMPDivSub64 ENDP
	;;
	;@This is a special version of the Multiply routine used in the modPow routine.
	;
	;This is basically the same alg as the Multiply routine except instead of creating a return array it uses the temp array from the modpow alg so that we dont keep creating and destroying objects.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AMPMultiply64 PROC mval1:PTR qword, mval2:PTR qword

		;This is a standard multiply alg, it creates a answer array the size of _v1size+_v2size+1 and clears all the cells
		;starting at cell 1 of _one of (stationary) the arrays the alg will multiply that cell by each cell in the other array and adding it to the (_currentcellv1-1)+(_currentcellv2-1)+1 in the answer array
		;at the end of a cycle it takes any carry from the last multiply add round and stores it in the next cell of the answer array
		;the last thing that is d_one is to check the high order bytes for _zero

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, qword ptr[r11]
		add		rax, qword ptr[r8]
		mov		qword ptr[r9], rax
		add		rax, 2
		mov		rcx, rax
		mov		rdi, r9
		add		rdi, 8
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf
		mov		rdi, 1
		mov		rsi, 1
		mov		rcx, 1
		mov		rdx, 0

MultLoop:

		mov		rbx, rdx
		mov		rax, qword ptr[r11+rdi*8]
		mul		qword ptr[r8+rsi*8]
		add		qword ptr[r9+rcx*8], rax
		adc		rdx, 0
		add		qword ptr[r9+rcx*8], rbx
		adc		rdx, 0
		inc		rcx
		inc		rdi
		cmp		rdi, qword ptr[r11]
		jna		MultLoop
		mov		qword ptr[r9+rcx*8], rdx
		mov		rdx, 0
		mov		rdi, 1
		inc		rsi
		mov		rcx, rsi
		cmp		rsi, qword ptr[r8]
		jna		MultLoop
		jmp		GoOut

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rax, 0
		RET

GoOut:

		mov		rcx, r9
		CHECKFORZEROS
		mov		r9, rax

OutOfHere:

		mov		rax, r9
		mov		r9, r8
		mov		r8, rax

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	AMPMultiply64 ENDP
	;;
	;@This is a special version of the Square routine used in the modPow routine.
	;
	;This is basically the same alg as the Square routine except instead of creating a return array it uses the temp array from the modpow alg so that we dont keep creating and destroying objects.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AMPSquare64 PROC val1:PTR qword

		;thiis is a modified version of the multiply alg

		ENTERFUNC _zero, _one, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, qword ptr[r8]
		add		rax, rax
		inc		rax
		mov		qword ptr[r9], rax
		mov		rdi, qword ptr[r8]
		cmp		rdi, 1
		ja		Cont
		mov		rax, qword ptr[r8+8]
		cmp		rax, 0
		je		ReturnZero
		cmp		rax, 1
		je		ReturnOne
		mul		rax
		mov		qword ptr[r9+8], rax
		mov		qword ptr[r9], 1
		cmp		rdx, 0
		je		GoOut
		mov		qword ptr[r9+16], rdx
		mov		qword ptr[r9], 2
		jmp		GoOut

Cont:

		mov		rcx, qword ptr[r9]
		mov		rdi, r9
		add		rdi, 8
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf
		mov		rdi, 1
		mov		rsi, 1
		mov		rcx, 1
		mov		rdx, 0

MultLoop:

		mov		rbx, rdx
		mov		rax, qword ptr[r8+rdi*8]
		mul		qword ptr[r8+rsi*8]
		add		rax, rbx
		adc		rdx, 0
		add		qword ptr[r9+rcx*8], rax
		adc		rdx, 0
		inc		rdi
		inc		rcx
		cmp		rdi, qword ptr[r8]
		jna		MultLoop
		mov		qword ptr[r9+rcx*8], rdx
		mov		rdx, 0
		mov		rdi, 1
		inc		rsi
		mov		rcx, rsi
		cmp		rsi, qword ptr[r8]
		jna		MultLoop
		jmp		GoOut

ReturnZero:

		mov		qword ptr[r9], 1
		mov		qword ptr[r9+8], 0
		jmp		OutOfHere

ReturnOne:

		mov		qword ptr[r9], 1
		mov		qword ptr[r9+8], 1
		jmp		OutOfHere

GoOut:

		mov		rcx, r9
		CHECKFORZEROS
		mov		r9, rax

OutOfHere:

		mov		rax, r8
		mov		r8, r9
		mov		r9, rax

		EXITFUNC _zero, _one, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	AMPSquare64 ENDP
	;;
	;@This is a special version of the Mod routine used in the modPow routine.
	;
	;This is basically the same alg as the Mod routine except instead of creating a return array it uses the temp array from the modpow alg so that we dont keep creating and destroying objects.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AMPMod64 PROC dtemp:PTR qword, dval2:PTR qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _one, _zero, _one, _zero, _one, _one, _zero

		;start the alg by checking to see if _val1 or _val2 is 0 or 1 and if _one of them are take the appropriate action

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, 0
		mov		_var2, 0
		mov		_t2, 0
		mov		_var4, 0
		mov		_var6, rax
		mov		_var3, 0
		mov		_t3, 0
		mov		_var5, rbx
		mov		r12, r13
		mov		_t1, 1
		mov		rax, qword ptr[r12]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r10, rax
		mov		rdi, qword ptr[r12]
		mov		rcx, rdi
		inc		rcx
		mov		rsi, r12
		mov		rdi, r10
		pushf
		cld
rep		movsq
		popf
		mov		r12, r10
		cmp		qword ptr[r12], 0
		je		GoOut
		cmp		qword ptr[r12], 1
		ja		CheckV1
		cmp		qword ptr[r12+8], 0
		je		GoOut
		cmp		qword ptr[r12+8], 1
		je		ReturnZero
		cmp		qword ptr[r8], 2
		ja		GetRetLength
		jb		JustOne
		mov		rdx, qword ptr[r8+16]
		cmp		rdx, qword ptr[r12+8]
		jnb		GetRetLength
		mov		rax, qword ptr[r8+8]
		div		qword ptr[r12+8]
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], rdx
		jmp		GoOut

JustOne:

		;if they both only have _one element then just do the division and return the answer

		cmp		qword ptr[r8], 2
		ja		CheckV1
		cmp		qword ptr[r8], 1
		je		BothOne
		mov		rax, qword ptr[r8+16]
		cmp		rax, qword ptr[r12+8]
		jnb		CheckV1
		mov		rdx, qword ptr[r8+16]
		mov		rax, qword ptr[r8+8]
		div		qword ptr[r12+8]
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], rdx
		jmp		GoOut

BothOne:

		xor		rdx, rdx
		mov		rax, qword ptr[r8+8]
		div		qword ptr[r12+8]
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], rdx
		jmp		GoOut

CheckV1:

		;this is to make sure _val1 gets checked, this part can be hit because of a branch

		cmp		qword ptr[r8], 0
		je		ReturnZero
		cmp		qword ptr[r8], 1
		ja		NextCheck
		cmp		qword ptr[r8+8], 0
		je		ReturnZero

NextCheck:

		;make sure the length of _val1 is equal to or greater than the length of _val2

		mov		rdi, qword ptr[r8]
		cmp		rdi, qword ptr[r12]
		jb		GoOut
		ja		GetRetLength
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rsi, r8
		add		rsi, rax
		mov		rdi, r12
		add		rdi, rax
		pushf
		std
repe	cmpsq
		jb		PrepGoOut
		ja		PrepGetRetLength
		popf
		jmp		ReturnOne

PrepGoOut:

		popf
		jmp		GoOut

PrepGetRetLength:

		popf

GetRetLength:

		;count how many of the high digits are equal, this is to fix the problem where the division comes out to 1 but should be FFFFFFFFFFFFFFFF, the firstpass variable is used in cunjuction
		;because typically the size off the return variable will have to be reduced by _one on the first correction pass only as long as we take the step of making the most significant byte in _val2
		;64 bits long by shifting all the bytes in _val2 by however many bits are required to make the msb 64 bits

		mov		rdi, qword ptr[r8]
		mov		rsi, qword ptr[r12]

SameLoop:

		mov		rax, qword ptr[r8+rdi*8]
		cmp		rax, qword ptr[r12+rsi*8]
		jne		SameOut
		inc		_t3
		dec		rdi
		sub		rsi, 1
		jnz		SameLoop

SameOut:

		;computer the length of the return _val by adding all the _bits in _val1 and subtracting the _bits in _val2 (this part isnt perfect) and dividing by 64

		mov		rax, qword ptr[r8]
		bsr		rcx, qword ptr[r8+rax*8]
		dec		rax
		mul		_sixtyfour
		add		rcx, rax
		mov		rax, qword ptr[r12]
		bsr		rbx, qword ptr[r12+rax*8]
		dec		rax
		mul		_sixtyfour
		add		rbx, rax
		sub		rcx, rbx
		mov		rax, rcx
		inc		rax
		xor		rdx, rdx
		div		_sixtyfour
		cmp		rdx, 0
		je		NoInc
		inc		rax

NoInc:

		;get the memory for the return val

		mov		r15, rax
		mov		rdi, qword ptr[r12]
		bsr		rax, qword ptr[r12+rdi*8]
		mov		rcx, 63
		sub		rcx, rax
		cmp		rcx, 0
		je		MainLoop
		mov		_var3, rcx
		mov		rdx, rcx
		mov		rcx, r12
		DivBitShiftLeft64
		
KeepIt:
		
		mov		rcx, r8
		mov		rdx, _var3
		DivBitShiftLeft64

MainLoop:

		;this is the main loop, here we figure out if we need to fill just rax or if we need to fill rax, rdx

		xor		rdx, rdx
		mov		rdi, qword ptr[r8]
		mov		rsi, qword ptr[r12]
		mov		rax, qword ptr[r8+rdi*8]
		dec		rdi
		cmp		rdi, 0
		je		KeepThis
		cmp		rax, qword ptr[r12+rsi*8]
		jnb		KeepThis
		mov		rdx, rax
		mov		rax, qword ptr[r8+rdi*8]

KeepThis:

		;do the division and move the value into r14, check if _t3 greater then 0, if it is check if r14==1, if _t1==1 then the size of the return val will have to be adjusted
		;this is d_one in a separate section which is why it passes through here

		div		qword ptr[r12+rsi*8]
		mov		r14, rax
		cmp		_t3, 0
		je		DoItAgain
		cmp		r14, 1
		jne		DoItAgain
		cmp		_t1, 1
		je		DoItAgain
		sub		r14, 2
		dec		_t3

DoItAgain:

		;save the value in the return val then multiply val2 by the value in r14 and store the result in r9
		;then we do a quick check of the lengths of _val1 and _val2, if _val2 is larger then it _zeros the result just saved in the return val and shifts the r9 value to the right _one

		cmp		r14, 0
		je		NextRetByte
		mov		rcx, r12
		mov		rdx, r14
		mov		rax, r9
		MPDivMult64
		mov		rdi, qword ptr[r8]
		cmp		rdi, qword ptr[r9]
		jb		SomethingIsWrong
		mov		rax, r15
		dec		rax
		cmp		rax, 0
		je		DoLastCheck
		mov		rcx, r9
		mov		rdx, rax
		DivByteShiftLeft64

DoLastCheck:

		;here we start checking the values to make sure the value in r9 is less than or equal to val1

		mov		rbx, 0
		mov		rdi, qword ptr[r8]
		cmp		rdi, qword ptr[r9]
		ja		DoSub
		jb		SomethingIsWrong
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rsi, r8
		add		rsi, rax
		mov		rdi, r9
		add		rdi, rax
		pushf
		std
repe	cmpsq
		ja		PrepDoSub
		jb		PrepDecAndRedo
		popf
		jmp		GoOut

PrepDoSub:

		popf
		jmp		DoSub

PrepDecAndRedo:

		popf
		jmp		DecAndRedo

DoSub:

		;subtract the value in r9 from the value in val1

		mov		rcx, r8
		mov		rdx, r9
		MPDivSub64
		mov		rdi, qword ptr[r8]
		cmp		rdi, qword ptr[r12]
		ja		NextRetByte
		jb		GoOut

		;this is the final check, it makes sure that r14 was large enough if it wasnt it is supposed to inc r14 and doitagain(but it isnt running that way right now)

		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rsi, r8
		add		rsi, rax
		mov		rdi, r12
		add		rdi, rax
		pushf
		std
repe	cmpsq
		ja		PrepNextRetByte
		popf
		jmp		GoOut

PrepNextRetByte:

		popf
		jmp		NextRetByte

DecAndRedo:

		;if the value in r14 was too large we dec r14 and doitagain, this is also the area where we adjust the retval size and adjust r14 if we are looking at equal most significant bytes

		cmp		r14, 1
		ja		JustDec
		cmp		_t3, 0
		je		JustDec
		sub		r14, 2
		dec		_t3
		cmp		_t1, 0
		je		DoItAgain
		mov		_t1, 0
		dec		r15
		jmp		DoItAgain

JustDec:

		dec		r14
		jmp		DoItAgain

SomethingIsWrong:

		;something was wrong with how we calculated the length of the return val, so _zero out the return val, store r14 in the next element down, shift r9 to the right _one, and redo the check

		dec		r15
		cmp		r15, 0
		je		GoOut
		mov		rcx, r9
		mov		rdx, 1
		DivByteShiftRight64
		mov		r9, rax
		jmp		DoLastCheck

NextRetByte:

		;everything went ok so on to the next byte

		dec		r15
		cmp		r15, 0
		je		GoOut
		jmp		MainLoop

ReturnZero:

		;return _zero without doing the division

		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		jmp		OutOfHere

ReturnOne:

		;return _one without doing the division

		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 1
		jmp		OutOfHere

ReturnVal1:

		;return val1 without doing the division

		jmp		OutOfHere

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		;do our check for any leading _zeros and dec the element at position 0

		mov		rcx, r8
		CHECKFORZEROS
		mov		r8, rax

OutOfHere:

		cmp		_var3, 0
		je		DontShiftIt
		mov		rcx, r8
		mov		rdx, _var3
		DivBitShiftRight64
		mov		_var9, rax
		mov		rdi, r8
		mov		rsi, rax
		mov		rcx, qword ptr[rax]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rcx, _var9
		call	Free64

DontShiftIt:

		;exit the alg

		mov		_jfunc, r11
		mov		rcx, r12
		call	Free64
		mov		rax, _jfunc

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _one, _zero, _one, _zero, _one, _one, _zero

		RET

	AMPMod64 ENDP
	;;
	;@This is a special version of the BitSShiftLeft routine used in the ModPow routine.
	;
	;This is basically the same alg as the BitSShiftLeft routine except instead of creating a return array it uses the temp array from the ModPow alg so that we dont keep creating and destroying objects.
	;@returns an array.
	;@param rcx - The first array.
	;@param rdx - The second array.
	;;
	AMPBitShiftLeft64 PROC val:PTR qword, count:qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r8, rax
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, r12
		xor		rdx, rdx
		div		_sixtyfour
		mov		_t1, rax
		cmp		rdx, 0
		ja		DoBits
		mov		rcx, r11
		mov		rdx, rax
		mov		rax, r8
		DivByteShiftLeft64

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

DoBits:

		mov		_t2, rdx
		mov		rax, qword ptr[r11]
		add		rax, _t1
		inc		rax
		mov		_jfunc, rax
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		rax, _jfunc
		mov		qword ptr[r8], rax
		mov		rcx, rax
		mov		rdi, r8
		add		rdi, 8
		mov		rax, 0
		pushf
		cld
rep		movsq
		popf
		mov		rdi, 1
		mov		rsi, 1
		add		rsi, _t1
		mov		rbx, 64
		sub		rbx, _t2
		mov		rcx, _t2

MainLoop:

		mov		rax, qword ptr[r11+rdi*8]
		shl		rax, cl
		add		qword ptr[r8+rsi*8], rax
		mov		rax, qword ptr[r11+rdi*8]
		xchg	rbx, rcx
		shr		rax, cl
		inc		rdi
		inc		rsi
		mov		qword ptr[r8+rsi*8], rax
		xchg	rbx, rcx
		cmp		rdi, qword ptr[r11]
		jna		MainLoop
		mov		rcx, r8
		CHECKFORZEROS
		mov		r8, rax

GoOut:

		mov		rax, r8
		mov		rcx, r11
		mov		_jfunc, rax
		call	Free64
		mov		rax, _jfunc

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		ErrorOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

ErrorOut:

		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	AMPBitShiftLeft64 ENDP
	;;
	;@brief Modular exponentiation of an array.
	;
	;Standard modular expontiation alg
	;@returns an array.
	;@param rcx - The Base array.
	;@param rdx - The Exponent array.
	;@param r8 - The Modulus array.
	;;
	AModPow64 PROC base:PTR qword, exp:PTR qword, amod:PTR qword	

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_v3, r8
		mov		_var1, 0
		mov		_t1, 0
		mov		_t2, 0
		mov		_var2, 0
		mov		_t3, 0
		mov		r11, rcx
		mov		r12, rdx
		mov		r13, r8
		lea		r10, _bits
		mov		rdi, qword ptr[r13]
		bsr		rax, qword ptr[r13+rdi*8]
		mov		rcx, 63
		sub		rcx, rax
		mov		_t2, rcx
		mov		rax, qword ptr[r11]
		cmp		rax, qword ptr[r13]
		cmovb	rax, qword ptr[r13]
		add		rax, rax
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		cmp		r8, 0
		je		MallocError
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		cmp		qword ptr[r12], 0
		je		ReturnOneMod
		cmp		qword ptr[r12], 1
		ja		ContinueCheck
		cmp		qword ptr[r12+8], 0
		je		ReturnOneMod
		cmp		qword ptr[r12+8], 1
		je		ReturnV1Mod

ContinueCheck:

		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r13]
		jne		NotEq

CheckEqLoop:

		mov		rax, qword ptr[r11+rdi*8]
		cmp		rax, qword ptr[r13+rdi*8]
		jne		NotEq
		sub		rdi, 1
		jnz		CheckEqLoop
		jmp		ReturnZero

NotEq:

		cmp		qword ptr[r11], 0
		je		ReturnZero
		cmp		qword ptr[r11], 1
		ja		ContinueCheck2
		cmp		qword ptr[r11+8], 0
		je		ReturnZero
		cmp		qword ptr[r11+8], 1
		je		ReturnOneMod

ContinueCheck2:

		cmp		qword ptr[r13], 0
		je		GoOut
		cmp		qword ptr[r13], 1
		ja		ContinueOn
		cmp		qword ptr[r13+8], 0
		je		GoOut
		cmp		qword ptr[r13+8], 1
		je		ReturnZero

ContinueOn:

		mov		rax, qword ptr[r13]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		rsi, r13
		mov		rdi, rax
		mov		r13, rax
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		r13, rax
		mov		_var1, 1
		mov		rax, qword ptr[r11]
		cmp		rax, qword ptr[r13]
		cmovb	rax, qword ptr[r13]
		add		rax, rax
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		r9, rax
		mov		_t1, 1
		mov		r14, qword ptr[r12]
		cmp		r14, 0
		je		GoOut
		bsr		r15, qword ptr[r12+r14*8]
		mov		_t3, r15
		cmp		r15, 0
		ja		DecAndGo
		mov		r15, 64
		mov		_t3, r15
		cmp		r14, 1
		jna		GoOut
		sub		r14, 1

DecAndGo:

		sub		r15, 1
		mov		_t3, r15

MainLoop:

		mov		rcx, r8
		MPSquare64
		mov		rbx, _var2
		mov		rcx, r8
		mov		rdx, r13
		mov		rax, _t2
		MPMod64
		mov		_var2, 1
		mov		r15, _t3
		mov		rax, qword ptr[r12+r14*8]
		test	rax, qword ptr[r10+r15*8]
		jz		NoMult
		mov		rcx, r11
		mov		rdx, r8
		MPMultiply64
		mov		rbx, _var2
		mov		rcx, r8
		mov		rdx, r13
		mov		rax, _t2
		MPMod64

NoMult:

		cmp		r15, 0
		jna		NextByte
		dec		r15
		mov		_t3, r15
		jmp		MainLoop


NextByte:

		mov		r15, 63
		mov		_t3, r15
		cmp		r14, 1
		jna		GoOut
		dec		r14
		jmp		MainLoop

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax
	
GoOut:

		cmp		_t1, 0
		je		TempNotCreated
		mov		rcx, r9
		call	Free64

TempNotCreated:

		cmp		_var1, 0
		je		V3NotCreated
		mov		rcx, r13
		call	Free64

V3NotCreated:

		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero

		RET	

ReturnOneMod:

		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 1
		jmp		GoOut

ReturnV1Mod:

		mov		rcx, r8
		mov		rdx, r13
		mov		rax, _t2
		Mod64
		mov		rcx, r8
		mov		r8, rax
		jmp		GoOut

ReturnZero:

		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		jmp		GoOut

	AModPow64	ENDP
	;;
	;@brief Bit shift left of an array.
	;
	;This alg starts by dividing the amount to shift by 64, if rax > 0 then it calls the ByteShiftLeft routine first with rax moved into rcx for the bytes to move, then the value in rdx after the division
	;is used for how many bits to shift.
	;@returns an array.
	;@param rcx - The array to shift.
	;@param rdx - The qword how many bits to shift.
	;;
	ABitShiftLeft64 PROC val:PTR qword, count:qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		rax, rdx
		xor		rdx, rdx
		div		_sixtyfour
		mov		_t1, rax
		mov		_t2, rdx
		add		rax, qword ptr[r11]
		cmp		rdx, 0
		cmove	rcx, _zero
		cmovne	rcx, _one
		add		rax, rcx
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_jfunc, r11
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r11, _jfunc
		mov		r8, rax
		cmp		r8, 0
		je		MallocError
		mov		rcx, qword ptr[r11]
		inc		rcx
		mov		rdi, r8
		mov		rsi, r11
		mov		rax, _t1
		mul		_eight
		add		rdi, rax
		pushf
		cld
rep		movsq
		popf
		cmp		_t1, 0
		je		NoZeros
		mov		rcx, _t1
		mov		rdi, r8
		mov		rax, 0
		pushf
		cld
rep		stosq
		popf

NoZeros:

		mov		rdi, qword ptr[r8]
		inc		rdi
		mov		qword ptr[r8+rdi*8], 0
		mov		qword ptr[r8], rdi
		mov		rdi, qword ptr[r8]
		inc		rdi

FindByte:

		cmp		rdi, 1
		je		GoOut
		dec		rdi
		cmp		_t2, 0
		je		GoOut
		bsr		rbx, qword ptr[r8+rdi*8]
		jz		FindByte
		mov		rbx, _t2
		mov		rcx, 64
		sub		rcx, rbx
		mov		rsi, rdi
		inc		rdi

ShiftLoop:

		mov		rax, qword ptr[r8+rsi*8]
		shr		rax, cl
		add		qword ptr[r8+rdi*8], rax
		xchg	rbx, rcx
		sal		qword ptr[r8+rsi*8], cl
		xchg	rbx, rcx
		dec		rdi
		dec		rsi
		cmp		rsi, 0
		jne		ShiftLoop
		mov		rcx, r8
		CHECKFORZEROS
		mov		r8, rax
		jmp		GoOut
		
MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	ABitShiftLeft64 ENDP
	;;
	;@brief Bit shift right of an array.
	;
	;This alg starts by dividing the number of bits to shift by 64, if rax > 0 it calls the ByteShiftRight routine first, and shifts the amount in rax after the div, then it shifts the amount in rdx after
	;the div to the right zeros out all bytes shifted out.
	;@returns an array.
	;@param rcx - The array to shift.
	;@param rdx - The qword how many bits to shift.
	;;
	ABitShiftRight64 PROC val:PTR qword, shift:qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		rax, rdx
		xor		rdx, rdx
		div		_sixtyfour
		mov		_t1, rax
		mov		_t2, rdx
		cmp		rax, qword ptr[r11]
		jnb		ReturnZero
		mov		rax, qword ptr[r11]
		sub		rax, _t1
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_jfunc, r11
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r11, _jfunc
		mov		r8, rax
		mov		rbx, 1
		cmp		r8, 0
		je		MallocError
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		sub		rcx, _t1
		mov		qword ptr[r8], rcx
		inc		rcx
		mov		rax, _t1
		mul		_eight
		add		rsi, rax
		add		rdi, 8
		pushf
		cld
rep		movsq
		popf
		cmp		_t2, 0
		je		GoOut
		mov		rbx, 64
		mov		rcx, _t2
		sub		rbx, rcx
		mov		rsi, 1
		mov		rdi, 2

ShiftLoop:

		shr		qword ptr[r8+rsi*8], cl
		xchg	rbx, rcx
		cmp		rdi, qword ptr[r8]
		ja		GoOut
		mov		rax, qword ptr[r8+rdi*8]
		shl		rax, cl
		add		qword ptr[r8+rsi*8], rax
		xchg	rbx, rcx
		inc		rdi
		inc		rsi
		jmp		ShiftLoop

ReturnZero:

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		cmp		r8, 0
		je		MallocError
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		jmp		GoOut

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:
		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	ABitShiftRight64 ENDP
	;;
	;@brief Byte shift left of an array.
	;
	;Shifts all the bytes to the left, zeroing out any bytes shifted out.
	;@returns an array.
	;@param rcx - The array to shift.
	;@param rdx - The qword how many bytes to shift.
	;;
	AByteShiftLeft64 PROC val:PTR qword, shift:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		rax, qword ptr[r11]
		add		rax, shift
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_jfunc, r11
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r11, _jfunc
		mov		r8, rax
		cmp		r8, 0
		je		MallocError
		mov		rax, qword ptr[r11]
		add		rax, shift
		mov		qword ptr[r8], rax
		mov		rsi, r11
		add		rsi, 8
		mov		rdi, r8
		mov		rax, shift
		inc		rax
		mul		_eight
		add		rdi, rax
		mov		rcx, qword ptr[r11]
		pushf
		cld
rep		movsq
		popf
		mov		rcx, shift
		mov		rax, rcx
		mul		_eight
		mov		rdi, r8
		add		rdi, rax
		mov		rax, 0
		pushf
		std
rep		stosq
		popf
		jmp		GoOut

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rax, r8

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	AByteShiftLeft64 ENDP
	;;
	;@brief Byte shift right of an array.
	;
	;Shifts each byte to the right, zeroing out the bytes shifted out.
	;@returns an array.
	;@param rcx - The array to shift.
	;@param rdx - The qword how many bytes to shift.
	;;
	AByteShiftRight64 PROC val:PTR qword, shift:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		rax, shift
		cmp		rax, qword ptr[r11]
		jnb		ReturnZero
		mov		rax, qword ptr[11]
		sub		rax, shift
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		mov		_jfunc, r11
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r11, _jfunc
		mov		r8, rax
		cmp		r8, 0
		je		MallocError
		mov		rcx, qword ptr[r11]
		mov		rax, rcx
		mul		_eight
		mov		rsi, r11
		add		rsi, rax
		sub		rcx, shift
		mov		qword ptr[r8], rcx
		mov		rax, rcx
		mul		_eight
		mov		rdi, r8
		add		rdi, rax
		pushf
		std
rep		movsq
		popf
		jmp		GoOut

ReturnZero:

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		mov		r8, rax
		cmp		r8, 0
		je		MallocError
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		jmp		GoOut

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rax, r8

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	AByteShiftRight64 ENDP
	;;
	;@brief To string function for arrays.
	;
	;This routine simply repeatedly calls the DivRem alg on a copy of _val (so we dont change the original) with the passed in the _tradix as the divisor.  At each pass through the loop, it calls DivRem
	;uses the value returned as the Remainder as the index into an array that holds the ascii value for the radix used, and puts that value into the _retval.  After the DivRem it checks the results of
	;the Divide part for zero, if equal it exits, if not, the result of the divide is used as the new input and the loop repeats.
	;@returns a char array.
	;@param rcx - The array to convert to string.
	;@param rdx - The radix to use.
	;;
	AToString64 PROC _val:PTR qword, _tradix:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
			
		mov		_v1, rcx
		mov		_v2, rdx
		call	ACountDigits
		mov		_t1, rax				;_t1==digit count
		mov		rsi, _v1
		mov		rax, qword ptr[rsi]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		_var1, rax				;_var1=_temp for alg
		mov		rdi, rax
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		mov		rcx, _t1
		add		rcx, 10
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		_retval, rax			;_retval for alg
		mov		rdi, rax
		mov		rcx, _t1
		add		rcx, 5
		mov		rax, 0
rep		stosq
		popf
		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		_t2, rax				;_t2==_divrem ret
		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		_t4, rax				;_t4==remainder ret val
		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		MallocError
		mov		_var2, rax				;_var2==radix	
		mov		qword ptr[rax], 1
		mov		rcx, _v2
		mov		qword ptr[rax+8], rcx
		cmp		_v2, 2
		jne		Check8
		lea		rax, _tobase2
		jmp		CheckDone

Check8:

		cmp		_v2, 8
		jne		Check10
		lea		rax, _tobase8
		jmp		CheckDone

Check10:

		cmp		_v2, 10
		jne		Check16
		lea		rax, _tobase10
		jmp		CheckDone

Check16:

		cmp		_v2, 16
		jne		Check32
		lea		rax, _tobase16
		jmp		CheckDone

Check32:

		cmp		_v2, 32
		jne		Check64
		lea		rax, _tobase32
		jmp		CheckDone

Check64:

		cmp		_v2, 64
		jne		GoOut
		lea		rax, _tobase64

CheckDone:

		mov		_var3, rax				;_var3==ptr to string array for conversion
		cmp		_t1, 0
		ja		ItsGood
		mov		rax, rax

ItsGood:

		dec		_t1

GenerateLoop:

		mov		rcx, _var1
		mov		rdx, _var2
		DivRem64
		mov		rcx, _t2
		mov		_t2, rax 
		call	Free64
		mov		rcx, _var1
		call	Free64
		mov		rcx, _t4
		call	Free64
		mov		rcx, _t2
		mov		rax, qword ptr[rcx]
		mov		_var1, rax
		mov		rax, qword ptr[rcx+8]
		mov		_t4, rax
		mov		rcx, _var3
		mov		rdi, qword ptr[rax+8]
		mov		rax, qword ptr[rcx+rdi*8]
		mov		rcx, _retval
		mov		rdi, _t1
		mov		byte ptr [rcx+rdi], al
		cmp		_t1, 0
		je		GoOut
		dec		_t1
		mov		rcx, _var1
		cmp		qword ptr[rcx], 1
		ja		GenerateLoop
		cmp		qword ptr[rcx+8], 0
		ja		GenerateLoop

GoOut:

		mov		rcx, _v1
		cmp		qword ptr[rcx], 1
		je		ItsOne
		mov		rax, rax

ItsOne:

		cmp		qword ptr[rcx+8], 0
		je		ItsZero
		mov		rax, rax

ItsZero:

		mov		rcx, _t2
		call	Free64
		mov		rcx, _t4
		call	Free64
		mov		rcx, _var1
		call	Free64
		mov		rcx, _var2
		call	Free64
		mov		rax, _retval
		jmp		OutNow

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		mov		_r8, r8
		mov		_r9, r9
		mov		_r10, r10
		mov		_r11, r11
		call	GetMem64					;call our getmem routine
		mov		r11, _r11
		mov		r10, _r10
		mov		r9, _r9
		mov		r8, _r8
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

OutNow:

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	AToString64 ENDP
	;;
	;@brief Old ToString routine, not used anymore.
	;@returns a char array.
	;@param rcx - The array to convert to string.
	;@param rdx - The radix to use.
	;;
	AToString642 PROC _val:PTR qword, _tradix:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero
			
		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var4, 0
		mov		rcx, 32
		call	GetMem64
		mov		_t3, rax
		mov		rcx, 32
		call	GetMem64
		mov		_var1, rax
		mov		rcx, 32
		call	GetMem64
		mov		qword ptr[rax], 1
		mov		rcx, _v2
		mov		qword ptr[rax+8], rcx
		mov		_t2, rax
		mov		rcx, 32
		call	GetMem64
		mov		qword ptr[rax], 1
		mov		_var2, 0
		mov		r11, _v1
		mov		rax, qword ptr[r11]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		_t1, rax
		mov		rsi, _v1
		mov		rcx, qword ptr[rsi]
		inc		rcx
		mov		rdi, rax
		pushf
		cld
rep		movsq
		popf
		mov		rcx, _v1
		mov		rax, qword ptr[rcx]
		mul		_twenty
		mov		_var2, rax
		cmp		rax, 0
		cmove	rax, _one
		mov		_var2, rax
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		_retval, rax
		mov		rcx, _var2
		mov		rdi, rax
		mov		rax, 0
		pushf
		cld
rep		stosb
		popf
		cmp		_v2, 2
		jne		Check8
		lea		rax, _tobase2
		jmp		CheckDone

Check8:

		cmp		_v2, 8
		jne		Check10
		lea		rax, _tobase8
		jmp		CheckDone

Check10:

		cmp		_v2, 10
		jne		Check16
		lea		rax, _tobase10
		jmp		CheckDone

Check16:

		cmp		_v2, 16
		jne		Check32
		lea		rax, _tobase16
		jmp		CheckDone

Check32:

		cmp		_v2, 32
		jne		Check64
		lea		rax, _tobase32
		jmp		CheckDone

Check64:

		cmp		_v2, 64
		jne		OutNow
		lea		rax, _tobase64

CheckDone:

		mov		_var3, rax

GenerateLoop:

		mov		rcx, _t1
		mov		rdx, _t2
		DivRem64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64
		mov		rcx, _t1
		call	Free64
		mov		r11, _t3
		mov		r11, qword ptr[r11+8]
		mov		rdi, qword ptr[r11+8]
		mov		r12, _var3
		mov		rax, qword ptr[r12+rdi*8]
		mov		r8, _retval
		mov		rdi, _var4
		mov		byte ptr[r8+rdi], al
		inc		rdi
		mov		_var4, rdi
		mov		r11, _t1
		cmp		qword ptr[r11], 1
		ja		GenerateLoop
		cmp		qword ptr[r11+8], 0
		ja		GenerateLoop
		mov		rdi, _var4
		mov		r8, _retval
		mov		byte ptr[r8+rdi], 0
		inc		rdi
		mov		byte ptr[r8+rdi], 0
		mov		rcx, _t1
		call	Free64
		mov		rcx, _t3
		call	Free64
		mov		rcx, _v1
		mov		rax, qword ptr[rcx]
		mul		_twenty
		mul		_eight
		add		rax, 5
		mov		rcx, rax
		call	GetMem64
		mov		_t3, rax
		mov		rcx, _v1
		mov		rax, qword ptr[rcx]
		mul		_twenty
		add		rax, 5
		mov		rcx, rax
		mov		rax, 0
		mov		rdi, _t3
		pushf
		cld
rep		stosb
		popf
		mov		rdi, _retval
		mov		rcx, 2000
		mov		rax, 0
		pushf
		cld
repne	scasb
		popf
		sub		rdi, _retval

FoundEnd:

		cmp		rdi, 1
		ja		Go
		cmp		rdi, 0
		je		Go2
		dec		rdi
		jmp		Go2

Go:

		dec		rdi

Go2:

		mov		rsi, rdi
		mov		rdi, 0
		xor		rcx, rcx
		mov		rax, _t3
		mov		r8, _retval
		mov		_var5, 1

ReverseLoop:

		mov		cl, byte ptr[r8+rsi]
		cmp		_var5, 0
		je		SkipIt
		cmp		cl, 0
		je		NextUn
		mov		_var5, 0

SkipIt:

		mov		byte ptr[rax+rdi], cl
		inc		rdi

NextUn:

		cmp		rsi, 0
		je		ROut
		dec		rsi
		jmp		ReverseLoop

ROut:

		mov		byte ptr[rax+rdi], 0
		inc		rdi
		mov		byte ptr[rax+rdi], 0
		mov		rcx, _retval
		mov		_retval, rdi
		call	Free64
		jmp		OutNow

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		OutNow
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

OutNow:

		mov		rax, _t3
		mov		rcx, rdi

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	AToString642 ENDP
	
		

	
	
	
	
	





	
	; Custom Build Step, including a listing file placed in intermediate directory
; but without Source Browser information
; debug:
; ml -c -Zi "-Fl$(IntDir)\$(InputName).lst" "-Fo$(IntDir)\$(InputName).obj" "$(InputPath)"
; release:
; ml -c "-Fl$(IntDir)\$(InputName).lst" "-Fo$(IntDir)\$(InputName).obj" "$(InputPath)"
; outputs:
; $(IntDir)\$(InputName).obj

; Custom Build Step, including a listing file placed in intermediate directory
; and Source Browser information also placed in intermediate directory
; debug:
; ml -c -Zi "-Fl$(IntDir)\$(InputName).lst" "-FR$(IntDir)\$(InputName).sbr" "-Fo$(IntDir)\$(InputName).obj" "$(InputPath)"
; release:
; ml -c "-Fl$(IntDir)\$(InputName).lst" "-FR$(IntDir)\$(InputName).sbr" "-Fo$(IntDir)\$(InputName).obj" "$(InputPath)"
; outputs:
; $(IntDir)\$(InputName).obj
; $(IntDir)\$(InputName).sbr
;define PORTABLE64BIT 1

	
	
	
	
	
		
	
	
	
	
	
	
	
	
	
		
	end
		;r8=retval
		;r9=dtemp
		;r10=dmult
		;r11=dval1
		;r12=dval2
		;r13=mylen
		;r14=car
		;r15=nextroundlength


